(*==============*)
        // AN EVENT IS RAN
(*--------------*)
        // A MESSAGE IS POSTED OR SENT
(*++++++++++++++*)
        // AN EXCEPTION IS RAISED
(*><><><><><><><*)
        // A THREADSAFE SECTION




// v4.0
//##############################################################################
//#################               ROUTINE                     ##################
//##############################################################################


        (*
          DEFINE THREADSAFE TO MAKE ALL THE OBJECTS THREADSAFE.
          || TAList
          || TAStringList
          || TARectList
        *)
        {$DEFINE LIST_THREADSAFE}
        {$DEFINE STRINGLIST_THREADSAFE}
        {$DEFINE RECTLIST_THREADSAFE}

        (*
          DEFINE EXCEPTION_LOG TO LOG EXCEPTION RAISES AND CATCHES.
        *)
        {$DEFINE EXCEPTION_LOG}

        (*
          DEFINE EXCEPTION_SHOWWINDOW SHOW A WINDOW WHEN AN EXCEPTION IS CAUGHT.
        *)
        {$DEFINE EXCEPTION_SHOWWINDOW}

        (*
          DEFINE EXCEPTION_FROZENLOCK TO RAISE AN EXCEPTION IF A LOCK IS FROZEN,
          THUS FREEZING A THREAD AN POSSIBLY THE APPLICATION.
        *)
        { $DEFINE EXCEPTION_FROZENLOCK}

        (*
          DEFINE EXCEPTION_DEBUG TO ALLOW THE APPLICATION TO GATHER LAST MINUTE
          DEBUG INFO ON ALL ASPECTS THAT COULD HELP DETECTING THE CAUSE OF THE
          EXCEPTION ERROR. THIS MAY SLOW DOWN THE APPLICATION.
        *)
        { $DEFINE EXCEPTION_DEBUG}


//##############################################################################
//#################               ROUTINE                     ##################
//##############################################################################
Unit ARoutine;




        		//###############
        		//## INTERFACE ##
        		//###############
        		Interface

        		//##############
        		//##   USES   ##
        		//##############
        		Uses
                        //######################################################
        		//######################## SYSTEM ######################
                                		  {System,}
                                		  HWindows,
                                                  HMessages,
                                                  HMultimedia,
                                                  HNetwork,

                        //######################################################
                        //######################### OWN ########################
                                                  AHeap;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#######################                                #######################
//#######################                                #######################
//##                        RECT, POINT, ARECT, STRING                        ##
//##                             CHAR, CONVERSION                             ##
//#######################                                #######################
//#######################                                #######################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




                        //##############
                        //### CONST  ###
                        //##############
                        Const
                        //######################################################
                        //################### RECT CONSTANTS ###################

  EMPTY_RECT: TRect   = (Left: 0; Top: 0; Right: 0; Bottom: 0);
  EMPTY_ARECT: TARect = (Left: 0; Top: 0; Width: 0; Height: 0);

                        //######################################################
                        //################### POINT CONSTANTS ##################

  EMPTY_POINT: TPoint = (X: 0; Y: 0);

                        //######################################################
                        //################### END OF LINE ######################

  EOL: String = #13#10;

                        //######################################################
                        //################## NUMERIC CONSTANTS #################

  MAX_BYTE    = 255;
  //#
  MAX_WORD    = 65536;
  MAX_CARD    = 4294967296;
  //#
  MAX_INT     = 2147483647;
  MAX_INT64   = 9223372036854775807;

                        //######################################################
                        //################### WINDOWS VERSION ##################

  WINDOWS_UNKNOWN       = 0;
  WINDOWS_31            = 1;
  WINDOWS_95            = 2;
  WINDOWS_98            = 3;
  WINDOWS_98SE          = 4;
  WINDOWS_ME            = 5;
  WINDOWS_NT            = 6;
  WINDOWS_2000          = 7;
  WINDOWS_XP            = 8;

                        //######################################################
                        //################### WINDOWS FAMILY ###################

  WINFAMILY_UNKNOWN     = 0;
  WINFAMILY_3X          = 1;
  WINFAMILY_9X          = 2;
  WINFAMILY_NT          = 3;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                            EXPLICIT INTERFACE                            ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                              init and final                              ##
//##############################################################################

procedure Init;
procedure Final;

//##############################################################################
//##                              freenil support                             ##
//##############################################################################

procedure FreeNil(var AObject);

//##############################################################################
//##                             point support                                ##
//##############################################################################

function NewPoint(AX, AY: Integer): TPoint;
function PointEqual(APoint1, APoint2: TPoint): Boolean; overload;
function PointEqual(AX, AY, BX, BY: Integer): Boolean; overload;
function PointEmpty(APoint: TPoint): Boolean;
//#
function PointInRect(APoint: TPoint; ARect: TRect): Boolean; overload;
function PointInRect(APoint: TPoint; ALeft, ATop, ARight, ABottom: Integer): Boolean; overload;
function PointInRect(X, Y: Integer; ARect: TRect): Boolean; overload;
function PointInRect(X, Y: Integer; ALeft, ATop, ARight, ABottom: Integer): Boolean; overload;
//#
function PointInARect(APoint: TPoint; ARect: TARect): Boolean; overload;
function PointInARect(APoint: TPoint; ALeft, ATop, AWidth, AHeight: Integer): Boolean; overload;
function PointInARect(X, Y: Integer; ARect: TARect): Boolean; overload;
function PointInARect(X, Y: Integer; ALeft, ATop, AWidth, AHeight: Integer): Boolean; overload;

//##############################################################################
//##                          point offset functions                          ##
//##############################################################################

procedure PointOffset(var Point: TPoint; Transl: Integer); overload;
procedure PointOffset(var Point: TPoint; Transl: TPoint); overload;
function PointGetOffset(Point1, Point2: TPoint): TPoint; overload;
function PointGetOffset(Point: TPoint; Rect: TRect): TPoint; overload;

//##############################################################################
//##                              rect support                                ##
//##############################################################################

function NewRect(ALeft, ATop, ARight, ABottom: Integer): TRect; overload;
function NewRect(ALeftTop, ARightBottom: TPoint): TRect; overload;
function NewRect(ARect: TARect): TRect; overload;
function RectEqual(ARect1, ARect2: TRect): Boolean;
function RectEmpty(ARect: TRect): Boolean;
//#
function RectInRect(ARect, ABounds: TRect): Boolean;

//##############################################################################
//##                          rect offset functions                           ##
//##############################################################################

procedure RectOffset(var Rect: TRect; Transl: TPoint); overload;
procedure RectOffset(var Rect: TRect; Transl: TRect); overload;
function RectGetOffset(Rect1, Rect2: TRect): TPoint; overload;
function RectGetOffset(Rect: TRect; Point: TPoint): TPoint; overload;

//##############################################################################
//##                              arect support                               ##
//##############################################################################

function NewARect(ALeft, ATop, AWidth, AHeight: Integer): TARect; overload;
function NewARect(ARect: TRect): TARect; overload;
function NewARect(ALeftTop, AWidthHeight: TPoint): TARect; overload;
function ARectEqual(ARect1, ARect2: TARect): Boolean;
function ARectEmpty(ARect: TARect): Boolean;
//#
function ARectInARect(ARect, ABounds: TARect): Boolean;

//##############################################################################
//##                          string match and replace                        ##
//##############################################################################

function StringMatchCount(S, Sub: String; CaseSensitive: Boolean = True): Integer; overload;
function StringMatchCount(S: String; Sub: Char; CaseSensitive: Boolean = True): Integer; overload;
//#
function StringMatchIndexLeft(S, Sub: String; CaseSensitive: Boolean = True): Integer; overload;
function StringMatchIndexLeft(S: String; Sub: Char; CaseSensitive: Boolean = True): Integer; overload;
function StringMatchIndexRight(S, Sub: String; CaseSensitive: Boolean = True): Integer; overload;
function StringMatchIndexRight(S: String; Sub: Char; CaseSensitive: Boolean = True): Integer; overload;
//#
function StringReplace(var S: String; const From, ReplTo: String; CaseSensitive: Boolean = True): Integer; overload;
function StringReplace(var S: String; From, ReplTo: Char; CaseSensitive: Boolean = True): Integer; overload;
function StringReplaceLeft(var S: String; const From, ReplTo: String; CaseSensitive: Boolean = True): Boolean; overload;
function StringReplaceLeft(var S: String; From, ReplTo: Char; CaseSensitive: Boolean = True): Boolean; overload;
function StringReplaceRight(var S: String; const From, ReplTo: String; CaseSensitive: Boolean = True): Boolean; overload;
function StringReplaceRight(var S: String; From, ReplTo: Char; CaseSensitive: Boolean = True): Boolean; overload;

//##############################################################################
//##                          string support                                  ##
//##############################################################################

function StringUpper(const S: String): String;
function StringLower(const S: String): String;
//#
function StringIn(const S: String; const A: array of String; CaseSensitive: Boolean = True): Boolean;
//#
function StringIsCard(const S: String): Boolean;
function StringIsInt(const S: String): Boolean;
function StringIsHex(const S: String): Boolean;
function StringIsExt(const S: String): Boolean;
function StringIsBool(const S: String): Boolean;
//#
function StringEqual(const S1, S2: String; CaseSensitive: Boolean = True): Boolean;
function StringCompare(const S1, S2: String): Integer;
//#
function StringTrim(const S: String): String;
function StringTrimLeft(const S: String): String;
function StringTrimRight(const S: String): String;
function StringTrimEllipse(const S: String; ALength: Integer): String;
//#
function StringFill(const AFill: String; ALength: Integer): String; overload;
function StringFill(const AFill: Char; ALength: Integer): String; overload;

//##############################################################################
//##                           char support                                   ##
//##############################################################################

function CharUpper(C: Char): Char;
function CharLower(C: Char): Char;
//#
function CharIn(C: Char; const A: array of Char; CaseSensitive: Boolean = True): Boolean; overload;
//#
function CharIsInt(C: Char): Boolean;
function CharIsHex(C: Char): Boolean;
//
function CharEqual(C1, C2: Char; CaseSensitive: Boolean = True): Boolean;
function CharCompare(C1, C2: Char): Integer;

//##############################################################################
//##                           pchar support                                  ##
//##############################################################################

// PCharLen returns the number of characters in Str, not
// counting the null terminator.
function PCharLen(const Str: PChar): Integer;

// PCharEnd returns a pointer to the null character that terminates Str.
function PCharEnd(const Str: PChar): PChar;

// PCharMove copies exactly Count characters from Source to Dest and
// returns Dest. Source and Dest may overlap.
function PCharMove(Dest: PChar; const Source: PChar; Count: Integer): PChar;

// PCharCopy copies Source to Dest and returns Dest.
function PCharCopy(Dest: PChar; const Source: PChar): PChar;

// PCharComp compares Str1 to Str2. The return value is less than 0 if
// Str1 < Str2, 0 if Str1 = Str2, or greater than 0 if Str1 > Str2.
function PCharComp(const Str1, Str2: PChar): Integer;

// PCharScan returns a pointer to the first occurrence of Chr in Str. If
// Chr does not occur in Str, PCharScan returns NIL. The null terminator
// is considered to be part of the string.
function PCharScan(const Str: PChar; Chr: Char): PChar;

// PCharRScan returns a pointer to the last occurrence of Chr in Str. If
// Chr does not occur in Str, PCharRScan returns NIL. The null terminator
// is considered to be part of the string.
function PCharRScan(const Str: PChar; Chr: Char): PChar;

// PCharPos returns a pointer to the first occurrence of Str2 in Str1. If
// Str2 does not occur in Str1, PCharPos returns NIL.
function PCharPos(const Str1, Str2: PChar): PChar;

// PCharUpper returns the uppercase value of Str and PCharLower returns
// the lowercase value of Str.
function PCharUpper(Str: PChar): PChar;
function PCharLower(Str: PChar): PChar;

//##############################################################################
//##                           convert support                                ##
//##############################################################################

function CartToHex(Value: Cardinal): String;
function IntToHex(Value: Integer): String; overload;
//#
function HexToCard(const S: String): Cardinal;
function HexToInt(const S: String): Integer;
//#
function CardToStr(Value: Cardinal): String;
function IntToStr(Value: Integer): String; overload;
function ExtToStr(Value: Extended; Decimals: Integer): String;
function BoolToStr(Value: Boolean): String;
//#
function StrToCard(const S: String): Cardinal;
function StrToInt(const S: String): Integer;
function StrToExt(const S: String): Extended;
function StrToBool(const S: String): Boolean;

//##############################################################################
//##                             numeric support                              ##
//##############################################################################

procedure Restrict(var Nr: Integer; Min, Max: Integer); overload;
procedure Restrict(var Nr: Cardinal; Min, Max: Cardinal); overload;
procedure Restrict(var Nr: Extended; Min, Max: Extended); overload;
procedure Restrict(var APoint: TPoint; ABounds: TRect); overload;
procedure Restrict(var ARect: TRect; ABounds: TRect); overload;
//#
function Ceil(X: Extended): Integer;
function Floor(X: Extended): Integer;
//#
function Min(X, Y: Integer): Integer; overload;
function Min(X, Y: Cardinal): Cardinal; overload;
function Min(X, Y: Extended): Extended; overload;
//#
function Max(X, Y: Integer): Integer; overload;
function Max(X, Y: Cardinal): Cardinal; overload;
function Max(X, Y: Extended): Extended; overload;
//#
procedure Exchange(var X, Y: Integer); overload;
procedure Exchange(var X, Y: Cardinal); overload;
procedure Exchange(var X, Y: Extended); overload;
//#
function Abs(X: Integer): Integer; overload;
function Abs(X: Extended): Extended; overload;
//#
function Sgn(X: Integer): Integer; overload;
function Sgn(X: Extended): Integer; overload;
//#
function Digits(Nr: Integer): Integer; overload;
function Digits(Nr: Cardinal): Integer; overload;
//#
function Power(Base: Extended; Exponent: Integer): Extended;
//#
function LowHigh(ALow, AHigh: SmallInt): Integer;
function LowInt(I: Integer): SmallInt;
function HighInt(I: Integer): SmallInt;



//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#########################                            #########################
//#########################                            #########################
//##                         STREAMS, FILE PATH DRIVE                         ##
//##                          MANIPULATION FUNCTIONS                          ##
//#########################                            #########################
//#########################                            #########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




                        //##############
                        //### CONST  ###
                        //##############
                        Const
                        //######################################################
                        //################### FILE SEEK MODE ###################

  FM_SEEK_BEGIN    = FILE_BEGIN;                 // seek from beginning of file.
  FM_SEEK_CURRENT  = FILE_CURRENT;               // seek from current position.
  FM_SEEK_END      = FILE_END;                   // seek from end of file.

                        //######################################################
                        //################### FILE OPEN MODE ###################

  FM_READ       = GENERIC_READ;         // open the file for reading.
  FM_WRITE      = GENERIC_WRITE;        // open the file for writing.
  FM_READWRITE  = FM_READ or FM_WRITE;
  FM_EMPTY      = 1;                    // empty the file on open.


                        //######################################################
                        //################### FILE SHARE MODE ##################

  FM_SHARE_NONE         = 0;
  FM_SHARE_READ         = FILE_SHARE_READ;
  FM_SHARE_WRITE        = FILE_SHARE_WRITE;
  FM_SHARE_READWRITE    = FM_SHARE_READ or FM_SHARE_WRITE;

                        //######################################################
                        //################### FILE ACCESS MODE #################

  FM_ACCESS_UNKNOWN    = 0;                             // the type of acess on the file is unknown.
  FM_ACCESS_RANDOM     = FILE_FLAG_RANDOM_ACCESS;       // random reads will be performed on the file.
                                                        // windows will optimize the cache for this.
  FM_ACCESS_SEQUENTIAL = FILE_FLAG_SEQUENTIAL_SCAN;     // sequential reads will be performed on the file.
                                                        // windows will optimize the cache for this.
  FM_ACCESS_FLUSH      = FILE_FLAG_WRITE_THROUGH;       // no caching will take place when writing to the file.

                        //######################################################
                        //################### FILE ATTRIBUTES ##################

  FA_NORMAL     = FILE_ATTRIBUTE_NORMAL;
  FA_READONLY   = FILE_ATTRIBUTE_READONLY;
  FA_ARCHIVE    = FILE_ATTRIBUTE_ARCHIVE;
  FA_HIDDEN     = FILE_ATTRIBUTE_HIDDEN;
  FA_SYSTEM     = FILE_ATTRIBUTE_SYSTEM;
  //#
  FA_VOLUMEID   = 8;
  FA_DIRECTORY  = FILE_ATTRIBUTE_DIRECTORY;

                        //######################################################
                        //#################### DRIVE TYPES #####################

  DT_UNKNOWN    = 0;                   // unable to determine the drive type.
  DT_NOROOT     = DRIVE_NO_ROOT_DIR;   // the drive has no root directory.
  DT_FLOPPY     = DRIVE_REMOVABLE;
  DT_FIXED      = DRIVE_FIXED;
  DT_NETWORK    = DRIVE_REMOTE;
  DT_CDROM      = DRIVE_CDROM;
  DT_RAM        = DRIVE_RAMDISK;

                        //######################################################
                        //################# FIRST DRIVE LETTER #################

  FIRST_DRIVE = 'a';

                        //######################################################
                        //################# DATE/TIME HELPERS ##################

  Months: array[0..11] of String = (
                                    ('January'),
                                    ('February'),
                                    ('March'),
                                    ('April'),
                                    ('May'),
                                    ('June'),
                                    ('July'),
                                    ('August'),
                                    ('September'),
                                    ('October'),
                                    ('November'),
                                    ('December')
                                   );

  Days: array[0..6] of String = (
                                 ('Sunday'),
                                 ('Monday'),
                                 ('Tuesday'),
                                 ('Wednesday'),
                                 ('Thursday'),
                                 ('Friday'),
                                 ('Saturday')
                                );




                        //##############
                        //### TYPES  ###
                        //##############
                        Type
                        //######################################################
                        //############### TYPE CONVERSION RECORDS ##############

  WordRec = packed record
    Bytes: array [0..1] of Byte;
  end;


  LongRec = packed record
    case Integer of
      0: (Bytes: array [0..3] of Byte);
      1: (Words: array [0..1] of Word);
      2: (SmallInts: array [0..1] of SmallInt);
  end;


  Int64Rec = packed record
    case Integer of
      0: (Bytes: array [0..7] of Byte);
      1: (Words: array [0..3] of Word);
      2: (SmallInts: array [0..3] of SmallInt);
      3: (Cardinals: array [0..1] of Cardinal);
      4: (Integers: array [0..1] of Integer);
  end;

                        //######################################################
                        //############## SEARCH REC USED BY FIND* ##############

  TSearchRec = record
    CreateTime,
    AccessTime,
    WriteTime: TSystemTime;
    Size: Int64;
    //#
    Attr: Integer;
    Name: String;
    //#
    //#
    ExcludeAttr: Integer;       // reserved
    FindHandle: Cardinal;       // reserved
    FindData: TWin32FindData;   // reserved
  end;

                        //######################################################
                        //####################### CLASS TYPES ##################

  TAFileStream = class;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                            EXPLICIT INTERFACE                            ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                                FILESTREAM                                ##
//##############################################################################




  // OPENS A SPECIFIED FILE FOR EDITING. IF THE FILE DOESN'T EXIST THEN
  // THE STREAM CREATES THE FILE. THE FILE IS NOT EMPTIED UNLESS THE
  // FM_EMPTY FLAG IS SPECIFIED.

  TAFileStream = class(TObject)
  protected
    FHandle: Cardinal;
    FFileName: String;
(*============================================================================*)
    function GetSize: Int64;
    procedure SetSize(ASize: Int64);
    function GetPosition: Int64;
    procedure SetPosition(APosition: Int64);
(*============================================================================*)
  public
    //#
    //#
    // if the file doesn't exist the the stream creates it. a check should
    // be made with fileexists before running the stream creator.
    constructor Create(const AFileName: String; AOpenMode, AShareMode: Cardinal; AAcessMode: Cardinal = FM_ACCESS_UNKNOWN; AFileAttr: Cardinal = FA_NORMAL);
    destructor Destroy; override;
    //#
    //#
    function Read(var ABuffer; ACount: Integer): Integer;
    function Write(const ABuffer; ACount: Integer): Integer;
    procedure ReadBuffer(var ABuffer; ACount: Integer);
    procedure WriteBuffer(const ABuffer; ACount: Integer);
    //#
    function Seek(AOffset: Int64; AOrigin: Integer): Int64;
    procedure FlushCache;
    //#
    property FileName: String read FFileName;
    property Handle: Cardinal read FHandle write FHandle;
    property Size: Int64 read GetSize write SetSize;
    //#
    property Position: Int64 read GetPosition write SetPosition;
    function BOF: Boolean;
    function EOF: Boolean;
  end;




//##############################################################################
//##                           path manipulation                              ##
//##############################################################################

function PathAddSlash(const APath: String): String;
function PathRemoveSlash(const APath: String): String;
function PathExtractFileName(const APath: String): String;
function PathExtractDrive(const APath: String): Char;
function PathExtract(const APath: String): String;
function PathDeph(const APath: String): Integer;
function PathShort(const APath: String): String;
function PathLong(const APath: String): String;
// returns the full path (including drive name and current dir).
function PathComplete(const APath: String): String;

//##############################################################################
//##                           file manipulation                              ##
//##############################################################################

function FileExtractName(const AFileName: String): String;
function FileChangeName(const AFileName, AName: String): String;
function FileExtractExt(const AFileName: String): String;
function FileChangeExt(const AFileName, AExtension: String): String;
function FileExists(const AFileName: String): Boolean;
function FileSize(const AFileName: String): Int64;
function FileGetAttr(const AFileName: String): Integer;
function FileSetAttr(const AFileName: String; AAttr: Integer): Boolean;
//#
function FileOpen(const AFileName: String; AOpenMode, AShareMode: Cardinal; AAcessMode: Cardinal = FM_ACCESS_UNKNOWN; AFileAttr: Cardinal = FA_NORMAL): Integer;
function FileCreate(const AFileName: String; AOpenMode, AShareMode: Cardinal; AAcessMode: Cardinal = FM_ACCESS_UNKNOWN; AFileAttr: Cardinal = FA_NORMAL): Integer;
function FileRead(AHandle: Cardinal; var ABuffer; ACount: Integer): Integer;
function FileWrite(AHandle: Cardinal; const ABuffer; ACount: Integer): Integer;
function FileSeek(AHandle: Cardinal; AOffset: Int64; AOrigin: Integer): Int64;
procedure FileClose(AHandle: Cardinal);
//#
function FileCopy(const AName, ADestinationName: String; Overwrite: Boolean = False): Boolean;
function FileDelete(const AFileName: String): Boolean;
function FileRename(const AOldName, ANewName: String; Overwrite: Boolean = False): Boolean;
//#
function DirCreate(const ADir: String): Boolean;
function DirCopy(const ADir, ADestinationDir: String; Overwrite: Boolean = False): Boolean;
function DirDelete(const ADir: String): Boolean;
function DirRename(const AOldDir, ANewDir: String; Overwrite: Boolean = False): Boolean;
function DirExists(const ADir: String): Boolean;
function DirEmpty(const ADir: String): Boolean;
function DirGetCurrent: String;
function DirSetCurrent(const ADir: String): Boolean;

//##############################################################################
//##                         file search functions                            ##
//##############################################################################

// findfirst searches the directory specified by path for the first file that
// matches the file name implied by APath and excludes attributes specified by the
// AExcludeAttr parameter. the result is returned in the F parameter.
// the path parameter is the directory and file name mask, including wildcard
// characters. will exclude '.' , '..' and the volume id from the search.
function FindFirst(const APath: String; AExcludeAttr: Integer; var F: TSearchRec): Boolean;

// findnext returns the next entry that matches the name and attributes specified
// in a previous call to FindFirst. the search record must be one that was passed
// to findfirst. will exclude '.' , '..' and volume id from the search.
function FindNext(var F: TSearchRec): Boolean;

// findclose terminates a findfirst or findnext sequence and frees the resources
// allocated by the FindFirst function.
procedure FindClose(var F: TSearchRec);

//##############################################################################
//##                            drive functions                               ##
//##############################################################################

function DriveGetName(ADrive: Char): String;
procedure DriveSetName(ADrive: Char; const AName: String);
function DriveFree(ADrive: Char): Int64;
function DriveSize(ADrive: Char): Int64;
function DriveID(ADrive: Char): String;
function DriveSystem(ADrive: Char): String;
function DriveType(ADrive: Char): Integer;
function DriveCount: Integer;
function DriveString: String;
function DriveReady(ADrive: Char): Boolean;
//#
function DrivesBitmaskToString(ABitmask: Integer): String;

//##############################################################################
//##                          network functions                               ##
//##############################################################################

function NetworkAddDrive(const Resource: String; ADrive: Char; const Password: String = ''): Boolean;
function NetworkRemoveDrive(ADrive: Char; Force: Boolean = False): Boolean;
function NetworkUniversalName(ADrive: Char): String;
function NetworkComputerName: String;
function NetworkUserName: String;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##################                                        ####################
//##################                                        ####################
//##                  POINTER LIST, STRING LIST, RECT LIST                    ##
//##                    THREAD, THREADLIST, THREADMANAGER                     ##
//##################                                        ####################
//##################                                        ####################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




                        //##############
                        //### CONST  ###
                        //##############
                        Const
                        //######################################################
                        //################## RECT LIST ADD MODE ################

  RL_MODE_NONE  = 0;              // add the rect unconditionally.

  RL_MODE_AND   = 1;              // decomposes the rects in the list in regions and those
                                  // that don't overlap the given item are removed.

  RL_MODE_OR    = 2;              // decomposes the rects in the list in regions and those
                                  // that overlap the given rect are removed. then the rect
                                  // is added in the list.

  RL_MODE_DIFF  = 3;              // decomposes the rects in the list in regions and those
                                  // that overlap the given rect are removed.

                        //######################################################
                        //#################### LOCK LAST ACTION ################

  L_TRY   = 'Try';
  L_SET   = 'Set';
  L_CLEAR = 'Clear';


                        //######################################################
                        //#################### THREAD PRIORITY #################

  TP_IDLE               = THREAD_PRIORITY_IDLE;
  TP_LOWEST             = THREAD_PRIORITY_LOWEST;
  TP_LOWER              = THREAD_PRIORITY_BELOW_NORMAL;
  TP_NORMAL             = THREAD_PRIORITY_NORMAL;
  TP_HIGHER             = THREAD_PRIORITY_ABOVE_NORMAL;
  TP_HIGHEST            = THREAD_PRIORITY_HIGHEST;
  TP_REALTIME           = THREAD_PRIORITY_TIME_CRITICAL;
    TP_TIMECRITICAL     = TP_REALTIME;




                        //##############
                        //### TYPES  ###
                        //##############
                        Type
                        //######################################################
                        //###################### BASIC EVENT ###################

  TOnEvent = procedure(Sender: TObject);

                        //######################################################
                        //###################### LIST TYPES ####################

  PPointerArray = ^TPointerArray;
  TPointerArray = array[0..(MAX_INT div SizeOf(Pointer) - SizeOf(Pointer))] of Pointer;

                        //######################################################
                        //################### STRINGLIST TYPES #################

  PStrArray = ^TStrArray;
  TStrArray = array[0..(MAX_INT div SizeOf(Pointer) - SizeOf(Pointer))] of String;

                        //######################################################
                        //#################### METHOD TYPECAST #################

  TMethod = record
    Code, Data: Pointer;
  end;

                        //######################################################
                        //####################### CLASS TYPES ##################

  TAList = class;
  TAStringList = class;
  TARectList = class;
  //#
  TALock = class;
  TAThread = class;
  TAThreadManager = class;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                            EXPLICIT INTERFACE                            ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                                    LIST                                  ##
//##############################################################################




  TAList = class(TObject)
  protected
    FList: PPointerArray;
    //#
    FCount,
    FCapacity,
    FDelta: Integer;
    //#
    {$IFDEF LIST_THREADSAFE}
     FLock: TALock;
     FAllowLock: Boolean;
    {$ENDIF}
(*============================================================================*)
    function GetItem(AIndex: Integer): Pointer;
    procedure SetItem(AIndex: Integer; AItem: Pointer);
    procedure SetCapacity(ACapacity: Integer);
(*============================================================================*)
    procedure Grow;
(*============================================================================*)
  public
    //#
    //#  
    constructor Create(AAllowLock: Boolean = False);
    destructor Destroy; override;
    //#
    //#
    function Add(AItem: Pointer): Integer;
    procedure Delete(AIndex: Integer);
    //#
    function Find(AItem: Pointer): Integer;
    //#
    property Items[AIndex: Integer]: Pointer read GetItem write SetItem;
    property Count: Integer read FCount;
    property Capacity: Integer read FCapacity;
    procedure Clear;
    //#
    //#
    function InBounds(AIndex: Integer): Boolean;
    //#
    function Remove(AItem: Pointer): Integer;
    procedure Insert(AIndex: Integer; AItem: Pointer);
    procedure Exchange(AIndex1, AIndex2: Integer);
    procedure Move(AIndex, ANewIndex: Integer);
  end;




//##############################################################################
//##                                 STRINGLIST                               ##
//##############################################################################




  TAStringList = class(TObject)
  protected
    FList: PStrArray;
    //#
    FCount,
    FCapacity,
    FDelta: Integer;
    //#
    //#
    {$IFDEF STRINGLIST_THREADSAFE}
     FLock: TALock;
     FAllowLock: Boolean;
    {$ENDIF}
(*============================================================================*)
    function GetItem(AIndex: Integer): String;
    procedure SetItem(AIndex: Integer; const AItem: String);
    procedure SetCapacity(ACapacity: Integer);
(*============================================================================*)
    procedure Grow;
(*============================================================================*)
  public
    //#
    //#  
    constructor Create(AAllowLock: Boolean = False);
    destructor Destroy; override;
    //#
    //#
    function Load(const AFileName: String): Boolean;
    procedure Save(const AFileName: String);
    //#
    //#
    function Add(const AItem: String): Integer;
    procedure Delete(AIndex: Integer);
    procedure Clear;
    //#
    function Find(const AItem: String; ACaseSensitive: Boolean = True): Integer;
    procedure Sort(AStart: Integer = 0; AEnd: Integer = -1);
    //#
    property Items[AIndex: Integer]: String read GetItem write SetItem;
    property Count: Integer read FCount;
    property Capacity: Integer read FCapacity;
    //#
    //#
    function InBounds(AIndex: Integer): Boolean;
    //#
    function Remove(const AItem: String; ACaseSensitive: Boolean = True): Integer;
    procedure Insert(AIndex: Integer; const AItem: String);
    procedure Exchange(AIndex1, AIndex2: Integer);
    procedure Move(AIndex, ANewIndex: Integer);
  end;




//##############################################################################
//##                                  RECTLIST                                ##
//##############################################################################




  TARectList = class(TObject)
  protected
    FList: PPointerArray;
    //#
    FCount,
    FCapacity,
    FDelta: Integer;
    //#
    //#
    {$IFDEF RECTLIST_THREADSAFE}
     FLock: TALock;
     FAllowLock: Boolean;
    {$ENDIF}
(*============================================================================*)
    function GetItem(AIndex: Integer): TRect;
    procedure SetItem(AIndex: Integer; AItem: TRect);
    procedure SetCapacity(ACapacity: Integer);
(*============================================================================*)
    procedure Grow;
(*============================================================================*)
  public
    //#
    //#
    constructor Create(AAllowLock: Boolean = False);
    destructor Destroy; override;
    //#
    //#
    procedure Add(AItem: TRect); overload;
    procedure Add(AItem: TRect; AMode: Integer); overload;
    procedure Delete(AIndex: Integer);
    procedure Clear;
    //#
    property Items[AIndex: Integer]: TRect read GetItem write SetItem;
    property Count: Integer read FCount;
    property Capacity: Integer read FCapacity;
    //#
    //#
    procedure RemoveEmpty;
    function RestrictIn(var ARect1: TRect; ARect2: TRect): Boolean;
    function RestrictOut(var ARect1: TRect; ARect2: TRect): Boolean;
  end;




//##############################################################################
//##                                  LOCK                                    ##
//##############################################################################




  TALock = class
  protected
    FLock: TLock;
    //#
    FName: String;
    FOwnerId: Cardinal;
    FLockCount: Integer;
    //#
    //#
    {$IFDEF EXCEPTION_FROZENLOCK}
     // debug information.
     // used by the frozen thread detector.
     FLastOwnerId: Cardinal;
     FCounter: Integer;
    {$ENDIF}
(*============================================================================*)
  public
    //#
    //#  
    constructor Create(const AName: String);
    destructor Destroy; override;
    //#
    //#
    property Lock: TLock read FLock;
    //#
    property Name: String read FName;
    property OwnerId: Cardinal read FOwnerId;
    property LockCount: Integer read FLockCount;
  end;



//##############################################################################
//##                                 THREAD                                   ##
//##############################################################################




  // EACH THREAD MUST HAVE A UNIQUE NAME. IF NOT THEN AN EXCEPTION IS RAISED.
  // FOR FREEONTERMINATE THREADS, MAKE SURE YOU ENTER THE FreeOnTerminateLock
  // AND FINDTHREAD BY NAME RETURNS A VALID RESULT.

  TAThread = class(TObject)
  protected
    FHandle,
    FThreadId: Cardinal;
    FName: String;
    //#
    FTerminated,
    FFinished,
    FFreeOnTerminate: Boolean;
    //#
    FSuspendCount: Integer;
    //#
    //#
    {$IFDEF EXCEPTION_DEBUG}
     // debug information.
     // used by the exception handler to get last minute info.
     FLastLockSet,
     FLastLockClear,
     FLastLockTry,
     FLastLockAction: String;
    {$ENDIF}
(*============================================================================*)
    FOnDestroy,
    FOnTerminate: TOnEvent;
(*============================================================================*)
    function GetPriority: Integer;
    procedure SetPriority(APriority: Integer);
    function GetSuspended: Boolean;
    function GetExitCode: Cardinal;
(*============================================================================*)
    procedure Execute; virtual; abstract;
(*============================================================================*)
  public
    //#
    //#
    constructor Create(const AName: String; ACreateSuspended: Boolean; AFreeOnTerminate: Boolean = False; APriority: Integer = TP_NORMAL);
    destructor Destroy; override;
    //#
    //#
    function Resume: Integer;
    function Suspend: Integer;
    property Suspended: Boolean read GetSuspended;
    property SuspendCount: Integer read FSuspendCount;
    //#
    procedure Terminate(AResumeAll: Boolean = True);
    procedure WaitFor(AHasMessageQueue: Boolean = False);
    //#
    procedure Yield(ATime: Integer = 0);
    //#
    property Name: String read FName;
    property Handle: Cardinal read FHandle;
    property ThreadId: Cardinal read FThreadId;
    property ExitCode: Cardinal read GetExitCode;
    //#
    property Priority: Integer read GetPriority write SetPriority;
    property Terminated: Boolean read FTerminated;
    property Finished: Boolean read FFinished;
    property FreeOnTerminate: Boolean read FFreeOnTerminate;
(*============================================================================*)
    property OnDestroy: TOnEvent read FOnDestroy write FOnDestroy;
    property OnTerminate: TOnEvent read FOnTerminate write FOnTerminate;
  end;




//##############################################################################
//##                                THREADMANAGER                             ##
//##############################################################################




  TAThreadManager = class(TObject)
  protected
    FLockList: TAList;
    FLockCount: Integer;
    //#
    FThreadList: TAList;
    FThreadCount: Integer;
    //#
    //#
    {$IFDEF EXCEPTION_DEBUG}
     // debug information.
     // used by the exception handler to get last minute info on mainthread.
     FLastLockSet,
     FLastLockClear,
     FLastLockTry,
     FLastLockAction: String;
    {$ENDIF}
(*============================================================================*)
    function GetLock(AIndex: Integer): TALock;
    function GetThread(AIndex: Integer): TAThread;
    //#
    function GetRunningThreadId: Cardinal;
    function GetRunningThread: TAThread;
(*============================================================================*)
  public
    //#
    //#  
    constructor Create;
    destructor Destroy; override;
    //#
    //#
    // if no lock is specified then the global lock is set or cleared.
    function SetExclusive(ATimes: Integer = 1): Integer; overload;
    function SetExclusive(ALock: TALock; ATimes: Integer = 1): Integer; overload;
    function ClearExclusive(ATimes: Integer = 1): Integer; overload;
    function ClearExclusive(ALock: TALock; ATimes: Integer = 1): Integer; overload;
    //#
    property RunningThreadId: Cardinal read GetRunningThreadId;
    property RunningThread: TAThread read GetRunningThread;
    //#
    function AddThread(AThread: TAThread): Boolean;
    function RemoveThread(AThread: TAThread): Boolean;
    function FindThread(AThread: TAThread): Integer; overload;
    function FindThread(AThreadId: Cardinal): Integer; overload;
    function FindThread(const AName: String): Integer; overload;
    //#
    property Threads[AIndex: Integer]: TAThread read GetThread;
    property ThreadCount: Integer read FThreadCount;
    //#
    //#
    function AddLock(ALock: TALock): Boolean;
    function RemoveLock(ALock: TALock): Boolean;
    //#
    property Locks[AIndex: Integer]: TALock read GetLock;
    property LockCount: Integer read FLockCount;
  end;




//##############################################################################
//##                         default thread function                          ##
//##############################################################################

// EVERY THREAD CREATED WILL CALL THIS PROCEDURE (AND THE 32 BIT PARAMETER
// ALLOWED BY THREADS CALLBACK PROCEDURES IS A POINTER TO THE THREAD OBJECT).
// AN EXCEPTION HANDLER IS INCLUDED.

function ThreadProc(AThread: TAThread): Integer;




                        //##############
                        //###  VAR   ###
                        //##############
                        Var
                        //######################################################
                        //###################### GLOBAL USED ###################

  LockListLock: TLock;
  ThreadListLock: TLock;

  ThreadManager: TAThreadManager;

  GlobalLock: TALock;
  FreeOnTerminateLock: TALock;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#################################             ################################
//#################################             ################################
//##                                  VARIOUS                                 ##
//#################################             ################################
//#################################             ################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




                        //##############
                        //### TYPES  ###
                        //##############
                        Type
                        //######################################################
                        //#################### METHOD TO PROC ##################

  PMethodToProc = ^TMethodToProc;
  TMethodToProc = packed record
        PopEax: Byte;              // $58      pop EAX
        PushSelf: record           //          push self
              OpCode: Byte;        // $68
              Self: Pointer;       // self
        end;
        PushEax: Byte;             // $50      push EAX
        Jump: record               //          jmp [target]
              OpCode: Byte;        // $FF
              ModRm: Byte;         // $25
              PTarget: ^Pointer;   // @target
              Target: Pointer;     // @MethodAddr
        end;
  end;




                        //##############
                        //### CONST  ###
                        //##############
                        Const
                        //######################################################
                        //################## MEMORY REQUEST CODES ##############

  MEMORY_PERCENT        = 0;
  MEMORY_SIZE           = 1;
  MEMORY_FREE           = 2;
  MEMORY_SWAPSIZE       = 3;
  MEMORY_SWAPFREE       = 4;




//##############################################################################
//##                           system functions                               ##
//##############################################################################

function CpuSpeed: Extended;
function CpuMmx: Boolean;
procedure CpuEmms;

// converts a procedure or function to a method.
function ProcedureToMethod(Self: TObject; ProcAddr: Pointer): TMethod;

// converts a method to a procedure or function. this works only for stdcall
// methods. you should free the procedure pointer when not needed anymore.
function MethodToProcedure(Self: TObject; MethodAddr: Pointer): Pointer; overload;
function MethodToProcedure(Method: TMethod): Pointer; overload;

//##############################################################################
//##                            windows remote                                ##
//##############################################################################

procedure WindowsShutDown(Force: Boolean = False);
procedure WindowsReboot(Force: Boolean = False);
function WindowsVersion: Integer;
function WindowsFamily: Integer;
function WindowsMemory(MemRequest: Integer): Cardinal;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//###############################              #################################
//###############################              #################################
//##                               EXCEPTIONS                                 ##
//###############################              #################################
//###############################              #################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




                        //##############
                        //### CONST  ###
                        //##############
                        Const
                        //######################################################
                        //################### GETMESSAGE RESULT ################

  GM_SUCCESS    = 1;
  GM_TERMINATE  = 0;
  GM_ERROR      = -1;

                        //######################################################
                        //################## BUTTON COLOR CODES ################

  ButtonColors:  array[0..2] of Cardinal = ($FFFFFF, $EFD3C6, 0);
  ButtonBorders: array[0..2] of Cardinal = ($FFFFFF, $C66931, 0);
  TextColors:    array[0..2] of Cardinal = (0, 0, $FFFFFF);

  Texts: array[0..1] of String = ('   Close window', 'Close application');

                        //######################################################
                        //##################### BUTTON STATES ##################

  BS_UP     = 0;
  BS_HOVER  = 1;
  BS_DOWN   = 2;

                        //######################################################
                        //############# FATAL EXCEPTION ERROR CODES ############

  FE_OUTOFMEMORY        = 1;
  FE_INVALIDPOINTER     = 2;
  FE_DIVBYZERO          = 3;
  FE_RANGE              = 4;
  FE_INTOVERFLOW        = 5;
  FE_INVALIDOP          = 6;
  FE_ZERODIVIDE         = 7;
  FE_OVERFLOW           = 8;
  FE_UNDERFLOW          = 9;
  FE_INVALIDCAST        = 10;
  FE_ACCESSVIOLATION    = 11;
  FE_PRIVILEGE          = 12;
  FE_CONTROLBREAK       = 13;
  FE_STACKOVERFLOW      = 14;
  //#
  FE_VARIANT_INVALIDVARCAST     = 15;
  FE_VARIANT_INVALIDVAROP       = 16;
  FE_VARIANT_DISPATCHERROR      = 17;
  FE_VARIANT_VARARRAYCREATE     = 18;
  FE_VARIANT_VARNOTARRAY        = 19;
  FE_VARIANT_VARARRAYBOUNDS     = 20;
  FE_ASSERTION                  = 21;
  //#
  FE_EXTERNAL                   = 22;
  FE_INTFCAST                   = 23;
  FE_SAFECALL                   = 24;

                        //######################################################
                        //################# EXCEPTION MESSAGES #################

  EMessages: array[1..24] of String = (
             ('Not enough memory.'),
             ('Invalid pointer operation.'),
             ('Division by zero.'),
             ('Range check error.'),
             ('Integer overflow.'),
             ('Invalid floating point operation.'),
             ('Floating point division by zero.'),
             ('Floating point overflow.'),
             ('Floating point underflow.'),
             ('Invalid class typecast.'),

             ('--> MESSAGE IS BUILT IN THE EXCEPTION RAISER'),

             ('Privileged instruction.'),
             ('Control-Break hit.'),
             ('Stack overflow.'),
             ('Invalid variant type conversion.'),
             ('Invalid variant operation.'),
             ('Variant method calls not supported.'),
             ('Error creating variant array.'),
             ('Variant is not an array.'),
             ('Variant array index out of bounds.'),
             ('Assertion failed.'),
             ('External exception.'),
             ('Interface not supported.'),
             ('Exception in safecall method.')
             );




                        //##############
                        //### TYPES  ###
                        //##############
                        Type
                        //######################################################
                        //####################### CLASS TYPES ##################

  Exception = class;
  //#
  EHeapException = class;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                            EXPLICIT INTERFACE                            ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                                 EXCEPTION                                ##
//##############################################################################




  // IF EXCEPTION_SHOWWINDOW IS SPECIFIED A WINDOW WILL BE SHOWN AND ALL
  // EXCEPTIONS RAISED AFTER WILL BE QUEUED UNTIL THE WINDOW IS CLOSED, ELSE
  // THE APPLICATION HALTS. IF AT LEAST ONE LOCK IS ENTERED THEN THE EXCEPTION
  // IS CONSIDERED FATAL.

  Exception = class(TObject)
  protected
    FMessage,
    FLocation: String;
    FRaiser: TObject;
    //#
    FFatal: Boolean;
(*============================================================================*)
  public
    //#
    //#
    constructor Create(const AMessage, ALocation: String; ARaiser: TObject; AFatal: Boolean = False);
    //#
    //#
    property Message: String read FMessage write FMessage;
    property Location: String read FLocation write FLocation;
    property Raiser: TObject read FRaiser write FRaiser;
    //#
    // the exception window will display only an option to terminate the
    // application without running the FinalUnits procedure.
    property Fatal: Boolean read FFatal write FFatal;
  end;




//##############################################################################
//##                               HEAPEXCEPTION                              ##
//##############################################################################




  ExceptClass = class of Exception;




  EHeapException = class(Exception)
  protected
    FAllowFree: Boolean;
(*=====================================================================================*)
  public
    procedure FreeInstance; override;
  end;




//##############################################################################
//##                     windows exception handling                           ##
//##############################################################################

procedure RaiseException(const AMessage, ALocation: String; ARaiser: TObject; AFatal: Boolean = False);
procedure RaiseLastError(const ALocation: String; ARaiser: TObject; AFatal: Boolean = False);
procedure HandleException(const ALocation: String);
//#
function ExceptionMessage(AException: Exception; const AHandleLocation: String): String;
function DebugMessage(AException: Exception; const AHandleLocation: String): String;
function ErrorMessage(AErrorCode: Integer): String;
//#
procedure ExceptionHandler(ExceptObject: TObject; ExceptAddr: Pointer);
procedure ErrorHandler(ErrorCode: Byte; ErrorAddr: Pointer);
procedure AbstractErrorHandler;

//##############################################################################
//##                      hidden exception functions                          ##
//##############################################################################

// converts a runtime error into a delphi class.
function GetExceptionClass(P: PExceptionRecord): ExceptClass;
// converts a runtime error into a delphi class instance.
function GetExceptionObject(P: PExceptionRecord): Exception;
// converts huge errorcodes into zero range.
function MapException(P: PExceptionRecord): Integer;

//##############################################################################
//##                        exception log functions                           ##
//##############################################################################

procedure ExceptionLog(const AMessage: String);

//##############################################################################
//##                      exception window functions                          ##
//##############################################################################

// THE WINDOW WILL FAIL AND FREEZE THE APPLICATION ONLY IF THERE IS AN EXCEPTION
// THROWN FROM WITHIN THE THREADLISTLOCK OR THE LOCKLISTLOCK. IN GENERAL TRY TO
// RELEASE THE LOCK BEFORE RAISING AN EXCEPTION.

function ExceptionThread(APointer: Pointer): Cardinal;
function ExceptionWindow(AHandle: Cardinal; AMsg, WParam, LParam: Integer): Integer; stdcall;

//##############################################################################
//##                         frozen window functions                          ##
//##############################################################################

function FrozenThread(APointer: Pointer): Cardinal;




                        //##############
                        //###  VAR   ###
                        //##############
                        Var
                        //######################################################
                        //##################### GLOBAL USED ####################

  OutOfMemory: EHeapException;
  InvalidPointer: EHeapException;

  ExceptionLock: TLock;
  ExceptionFatal: Boolean = False;

  {$IFDEF EXCEPTION_LOG}
   ExceptionLogFile: TAFileStream;
  {$ENDIF}

  {$IFDEF EXCEPTION_SHOWWINDOW}
   ExceptionVisible: Boolean = False;
   ExceptionThreadTerminated: Boolean = False;
   ExceptionThreadFinished: Boolean = False;
   ExceptionThreadHandle: Cardinal;
   ExceptionThreadId: Cardinal;
   ExceptionWindowRect: TRect;
   ExceptionWindowHandle: Cardinal;
   ExceptionWindowDC: Cardinal;
   ExceptionMouseTrack: TTrackMouseEvent;
   ExceptionMsgHandle: Cardinal;
   ExceptionDebugHandle: Cardinal;
   ExceptionCloseRect: TRect;
   ExceptionDebugRect: TRect;
   ExceptionCloseState: Integer = BS_UP;
   ExceptionDebugState: Integer = BS_UP;
  {$ENDIF}

  {$IFDEF EXCEPTION_FROZENLOCK}
   FrozenThreadTerminated: Boolean = False;
   FrozenThreadFinished: Boolean = False;
   FrozenThreadHandle: Cardinal;
   FrozenThreadId: Cardinal;
  {$ENDIF}

                        //##############
                        //### CONST  ###
                        //##############
                        Const
                        //######################################################
                        //############# BASE CLASS FOR EXCEPTION ###############

   ExceptionWindowClass: TWndClass = (
             Style: CS_OWNDC;
             lpfnWndProc: @DefWindowProc;
             cbClsExtra: 0;
             cbWndExtra: 0;
             hInstance: 0;
             hIcon: 0;
             hCursor: 0;
             hbrBackground: 0;
             lpszMenuName: nil;
             lpszClassName: 'classException');




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                              IMPLEMENTATION                              ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
Implementation




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#######################                                #######################
//#######################                                #######################
//##                        RECT, POINT, ARECT, STRING                        ##
//##                             CHAR, CONVERSION                             ##
//#######################                                #######################
//#######################                                #######################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                              freenil support                             ##
//##############################################################################

procedure FreeNil(var AObject);
begin
        if (TObject(AObject) <> nil) then TObject(AObject).Free;
        TObject(AObject):= nil;
end;

//##############################################################################
//##                              point support                               ##
//##############################################################################

function NewPoint(AX, AY: Integer): TPoint;
begin
        Result.X:= AX;
        Result.Y:= AY;
end;


function PointEqual(APoint1, APoint2: TPoint): Boolean;
begin
        Result:= (APoint1.X = APoint2.X)and(APoint1.Y = APoint2.Y);
end;


function PointEqual(AX, AY, BX, BY: Integer): Boolean;
begin
        Result:= (AX = BX)and(AY = BY);
end;


function PointEmpty(APoint: TPoint): Boolean;
begin
        Result:= (APoint.X <= 0)or(APoint.Y <= 0);
end;


//#
function PointInRect(APoint: TPoint; ARect: TRect): Boolean;
begin
        Result:= (APoint.X >= ARect.Left)and(APoint.Y >= ARect.Top)and
                 (APoint.X < ARect.Right)and(APoint.Y < ARect.Bottom);
end;


function PointInRect(APoint: TPoint; ALeft, ATop, ARight, ABottom: Integer): Boolean;
begin
        Result:= (APoint.X >= ALeft)and(APoint.Y >= ATop)and(APoint.X < ARight)and(APoint.Y < ABottom);
end;


function PointInRect(X, Y: Integer; ARect: TRect): Boolean;
begin
        Result:= (X >= ARect.Left)and(Y >= ARect.Top)and(X < ARect.Right)and(Y < ARect.Bottom);
end;


function PointInRect(X, Y: Integer; ALeft, ATop, ARight, ABottom: Integer): Boolean;
begin
        Result:= (X >= ALeft)and(Y >= ATop)and(X < ARight)and(Y < ABottom);
end;


//#
function PointInARect(APoint: TPoint; ARect: TARect): Boolean;
begin
        Result:= (APoint.X >= ARect.Left)and(APoint.X >= ARect.Top)and
                 (APoint.X < ARect.Left + ARect.Width)and(APoint.Y < ARect.Top + ARect.Height);
end;


function PointInARect(APoint: TPoint; ALeft, ATop, AWidth, AHeight: Integer): Boolean;
begin
        Result:= (APoint.X >= ALeft)and(APoint.Y >= ATop)and(APoint.X < ALeft + AWidth)and(APoint.Y < ATop + AHeight);
end;


function PointInARect(X, Y: Integer; ARect: TARect): Boolean;
begin
        Result:= (X >= ARect.Left)and(Y >= ARect.Top)and(X < ARect.Left + ARect.Width)and(Y < ARect.Top + ARect.Height);
end;


function PointInARect(X, Y: Integer; ALeft, ATop, AWidth, AHeight: Integer): Boolean;
begin
        Result:= (X >= ALeft)and(Y >= ATop)and(X < ALeft + AWidth)and(Y < ATop + AHeight);
end;

//##############################################################################
//##                          point offset functions                          ##
//##############################################################################

procedure PointOffset(var Point: TPoint; Transl: Integer);
begin
        Inc(Point.X, Transl);
        Inc(Point.Y, Transl);
end;


procedure PointOffset(var Point: TPoint; Transl: TPoint);
begin
        Dec(Point.X, Transl.X);
        Dec(Point.Y, Transl.X);
end;


function PointGetOffset(Point1, Point2: TPoint): TPoint;
begin
        Result.X:= Point2.X - Point1.X;
        Result.Y:= Point2.Y - Point1.Y;
end;


function PointGetOffset(Point: TPoint; Rect: TRect): TPoint;
begin
        Result.X:= Point.X - Rect.Left;
        Result.Y:= Point.Y - Rect.Top;
end;

//##############################################################################
//##                              rect support                                ##
//##############################################################################

function NewRect(ALeft, ATop, ARight, ABottom: Integer): TRect;
begin
        Result.Left:= ALeft;
        Result.Top:= ATop;
        Result.Right:= ARight;
        Result.Bottom:= ABottom;
end;


function NewRect(ALeftTop, ARightBottom: TPoint): TRect;
begin
        Result.Left:= ALeftTop.X;
        Result.Top:= ALeftTop.Y;
        Result.Right:= ARightBottom.X;
        Result.Bottom:= ARightBottom.Y;
end;


function NewRect(ARect: TARect): TRect;
begin
        Result.Left:= ARect.Left;
        Result.Top:= ARect.Top;
        Result.Right:= ARect.Left + ARect.Width;
        Result.Bottom:= ARect.Top + ARect.Height;
end;


function RectEqual(ARect1, ARect2: TRect): Boolean;
begin
        Result:= (ARect1.Left = ARect2.Left)and(ARect1.Top = ARect2.Top)and
                 (ARect1.Right = ARect2.Right)and(ARect1.Bottom = ARect2.Bottom);
end;


function RectEmpty(ARect: TRect): Boolean;
begin
        Result:= (ARect.Right <= ARect.Left)or(ARect.Bottom <= ARect.Top);
end;


//#
function RectInRect(ARect, ABounds: TRect): Boolean;
begin
        Result:= (ARect.Left >= ABounds.Left)and(ARect.Top >= ABounds.Top)and
                 (ARect.Right < ABounds.Right)and(ARect.Bottom < ABounds.Bottom);
end;

//##############################################################################
//##                          rect offset functions                           ##
//##############################################################################

procedure RectOffset(var Rect: TRect; Transl: TPoint);
begin
        Rect.Left:= Rect.Left + Transl.X;
        Rect.Right:= Rect.Right + Transl.X;
        Rect.Top:= Rect.Top + Transl.Y;
        Rect.Bottom:= Rect.Bottom + Transl.Y;
end;


procedure RectOffset(var Rect: TRect; Transl: TRect);
begin
        Rect.Left:= Rect.Left + Transl.Left;
        Rect.Right:= Rect.Right + Transl.Top;
        Rect.Top:= Rect.Top + Transl.Right;
        Rect.Bottom:= Rect.Bottom + Transl.Bottom;
end;


function RectGetOffset(Rect1, Rect2: TRect): TPoint;
begin
        Result.X:= Rect2.Left - Rect1.Left;
        Result.Y:= Rect2.Top - Rect1.Top;
end;


function RectGetOffset(Rect: TRect; Point: TPoint): TPoint;
begin
        Result.X:= Rect.Left - Point.X;
        Result.Y:= Rect.Top - Point.Y;
end;

//##############################################################################
//##                              arect support                               ##
//##############################################################################

function NewARect(ALeft, ATop, AWidth, AHeight: Integer): TARect;
begin
        Result.Left:= ALeft;
        Result.Top:= ATop;
        Result.Width:= AWidth;
        Result.Height:= AHeight;
end;


function NewARect(ALeftTop, AWidthHeight: TPoint): TARect;
begin
        Result.Left:= ALeftTop.X;
        Result.Top:= ALeftTop.Y;
        Result.Width:= AWidthHeight.X;
        Result.Height:= AWidthHeight.Y;
end;


function NewARect(ARect: TRect): TARect;
begin
        Result.Left:= ARect.Left;
        Result.Top:= ARect.Top;
        Result.Width:= ARect.Right - ARect.Left;
        Result.Height:= ARect.Bottom - ARect.Top;
end;


function ARectEqual(ARect1, ARect2: TARect): Boolean;
begin
        Result:= (ARect1.Left = ARect2.Left)and(ARect1.Top = ARect2.Top)and
                 (ARect1.Width = ARect2.Width)and(ARect1.Height = ARect2.Height);
end;


function ARectEmpty(ARect: TARect): Boolean;
begin
        Result:= (ARect.Width <= 0)or(ARect.Height <= 0);
end;


//#
function ARectInARect(ARect, ABounds: TARect): Boolean;
begin
        Result:= (ARect.Left >= ABounds.Left)and(ARect.Top >= ABounds.Top)and
                 (ARect.Left + ARect.Width < ABounds.Left + ABounds.Width)and
                 (ARect.Top + ARect.Height < ABounds.Top + ABounds.Height);
end;

//##############################################################################
//##                          string match and replace                        ##
//##############################################################################

function StringMatchCount(S, Sub: String; CaseSensitive: Boolean = True): Integer;
var i, ASubLength, ALength: Integer;
begin
        Result:= 0;
        ASubLength:= Length(Sub);
        ALength:= Length(S);

        if (ASubLength > 0)and(ALength >= ASubLength) then
        begin
                case CaseSensitive of
                  True:
                  begin
                        for i:= 1 to ALength - ASubLength + 1 do
                        if (Sub = Copy(S, i, ASubLength)) then Inc(Result);
                  end;
                  //
                  else begin
                        Sub:= StringLower(Sub);
                        S:= StringLower(S);

                        for i:= 1 to ALength - ASubLength + 1 do
                        if (Sub = Copy(S, i, ASubLength)) then Inc(Result);
                  end;
                  //
                end;
        end;
end;


function StringMatchCount(S: String; Sub: Char; CaseSensitive: Boolean = True): Integer;
var i, ALength: Integer;
begin
        Result:= 0;
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                case CaseSensitive of
                  True:
                  begin
                        for i:= 1 to ALength do
                        if (S[i] = Sub) then Inc(Result);
                  end;
                  //
                  else begin
                        Sub:= CharLower(Sub);
                        S:= StringLower(S);

                        for i:= 1 to ALength do
                        if (S[i] = Sub) then Inc(Result);
                  end;
                  //
                end;
        end;
end;


function StringMatchIndexLeft(S, Sub: String; CaseSensitive: Boolean = True): Integer;
var ALength, ASubLength: Integer;
begin
        ASubLength:= Length(Sub);
        ALength:= Length(S);

        if (ASubLength > 0)and(ALength >= ASubLength) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for Result:= 1 to ALength - ASubLength + 1 do
                        if (Sub = Copy(S, Result, ASubLength)) then Exit;
                 end;
                 //
                 else begin
                        for Result:= 1 to ASubLength do
                        if (Byte(S[Result]) >= Byte('A'))and(Byte(S[Result]) <= Byte('Z')) then Inc(Byte(S[Result]), 32);

                        Sub:= StringLower(Sub);

                        for Result:= 1 to ALength - ASubLength + 1 do
                        begin
                                if (Byte(S[Result + ASubLength - 1]) >= Byte('A'))and(Byte(S[Result + ASubLength - 1]) <= Byte('Z')) then Inc(Byte(S[Result + ASubLength - 1]), 32);
                                if (Sub = Copy(S, Result, ASubLength)) then Exit;
                        end;
                 end;
                 //
                end;
        end;

        Result:= 0;
end;


function StringMatchIndexLeft(S: String; Sub: Char; CaseSensitive: Boolean = True): Integer;
var ALength: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for Result:= 1 to ALength do
                        if (Sub = S[Result]) then Exit;
                 end;
                 //
                 else begin
                        Sub:= CharLower(Sub);

                        for Result:= 1 to ALength do
                        begin
                                if (Byte(S[Result]) >= Byte('A'))and(Byte(S[Result]) <= Byte('Z')) then Inc(Byte(S[Result]), 32);
                                if (Sub = S[Result]) then Exit;
                        end;
                 end;
                 //
                end;
        end;

        Result:= 0;
end;


function StringMatchIndexRight(S, Sub: String; CaseSensitive: Boolean = True): Integer;
var ALength, ASubLength: Integer;
begin
        ASubLength:= Length(Sub);
        ALength:= Length(S);

        if (ASubLength > 0)and(ALength >= ASubLength) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for Result:= ALength - ASubLength + 1 downto 1 do
                        if (Sub = Copy(S, Result, ASubLength)) then Exit;
                 end;
                 //
                 else begin
                        for Result:= ALength downto ALength - ASubLength + 1 do
                        if (Byte(S[Result]) >= Byte('A'))and(Byte(S[Result]) <= Byte('Z')) then Inc(Byte(S[Result]), 32);

                        Sub:= StringLower(Sub);

                        for Result:= ALength - ASubLength + 1 downto 1 do
                        begin
                                if (Byte(S[Result]) >= Byte('A'))and(Byte(S[Result]) <= Byte('Z')) then Inc(Byte(S[Result]), 32);
                                if (Sub = Copy(S, Result, ASubLength)) then Exit;
                        end;
                 end;
                 //
                end;
        end;

        Result:= 0;
end;


function StringMatchIndexRight(S: String; Sub: Char; CaseSensitive: Boolean = True): Integer;
var ALength: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for Result:= ALength downto 1 do
                        if (Sub = S[Result]) then Exit;
                 end;
                 //
                 else begin
                        Sub:= CharLower(Sub);

                        for Result:= ALength downto 1 do
                        begin
                                if (Byte(S[Result]) >= Byte('A'))and(Byte(S[Result]) <= Byte('Z')) then Inc(Byte(S[Result]), 32);
                                if (Sub = S[Result]) then Exit;
                        end;
                 end;
                 //
                end;
        end;

        Result:= 0;
end;



//#
function StringReplace(var S: String; const From, ReplTo: String; CaseSensitive: Boolean = True): Integer;
var i, ALength, AFromLength, AReplToLength: Integer;
begin
        Result:= 0;
        ALength:= Length(S);
        AFromLength:= Length(From);
        AReplToLength:= Length(ReplTo);

        if (AFromLength > 0)and(ALength >= AFromLength) then
        begin
                i:= 1;

                while (i <= ALength - AFromLength + 1) do
                begin
                        if StringEqual(Copy(S, i, AFromLength), From, CaseSensitive) then
                        begin
                                S:= Copy(S, 1, i - 1) + ReplTo + Copy(S, i + AFromLength, MAX_INT);
                                Inc(Result);

                                ALength:= ALength - AFromLength + AReplToLength;
                                i:= i + AReplToLength;
                        end
                        else Inc(i);
                end;
        end;
end;


function StringReplace(var S: String; From, ReplTo: Char; CaseSensitive: Boolean = True): Integer;
var i, ALength: Integer;
    AC, AF: Byte;
begin
        Result:= 0;
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for i:= 1 to ALength do
                        if (S[i] = From) then
                        begin
                                S[i]:= From;
                                Inc(Result);
                        end;
                 end;
                 //
                 else begin
                        AF:= Byte(CharLower(From));

                        for i:= 1 to ALength do
                        begin
                                AC:= Byte(S[i]);
                                if (AC >= Byte('A'))and(AC <= Byte('Z')) then Inc(AC, 32);

                                if (AC = AF) then
                                begin
                                        S[i]:= From;
                                        Inc(Result);
                                end;
                        end;
                 end;
                 //
                end;
        end;
end;


function StringReplaceLeft(var S: String; const From, ReplTo: String; CaseSensitive: Boolean = True): Boolean;
var i: Integer;
begin
        Result:= False;
        i:= StringMatchIndexLeft(S, From, CaseSensitive);

        if (i > 0) then
        begin
                S:= Copy(S, 1, i - 1) + ReplTo + Copy(S, i + Length(From), MAX_INT);
                Result:= True;
        end;
end;


function StringReplaceLeft(var S: String; From, ReplTo: Char; CaseSensitive: Boolean = True): Boolean;
var i, ALength: Integer;
    AC, AF: Byte;
begin
        Result:= False;
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for i:= 1 to ALength do
                        if (S[i] = From) then
                        begin
                                S[i]:= From;
                                Result:= True;

                                Exit;
                        end;
                 end;
                 //
                 else begin
                        AF:= Byte(CharLower(From));

                        for i:= 1 to ALength do
                        begin
                                AC:= Byte(S[i]);
                                if (AC >= Byte('A'))and(AC <= Byte('Z')) then Inc(AC, 32);

                                if (AC = AF) then
                                begin
                                        S[i]:= From;
                                        Result:= True;

                                        Exit;
                                end;
                        end;
                 end;
                 //
                end;
        end;
end;


function StringReplaceRight(var S: String; const From, ReplTo: String; CaseSensitive: Boolean = True): Boolean;
var i: Integer;
begin
        Result:= False;
        i:= StringMatchIndexRight(S, From, CaseSensitive);

        if (i > 0) then
        begin
                S:= Copy(S, 1, i - 1) + ReplTo + Copy(S, i + Length(From), MAX_INT);
                Result:= True;
        end;
end;


function StringReplaceRight(var S: String; From, ReplTo: Char; CaseSensitive: Boolean = True): Boolean;
var i, ALength: Integer;
    AC, AF: Byte;
begin
        Result:= False;
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                case CaseSensitive of
                 True:
                 begin
                        for i:= ALength downto 1 do
                        if (S[i] = From) then
                        begin
                                S[i]:= From;
                                Result:= True;

                                Exit;
                        end;
                 end;
                 //
                 else begin
                        AF:= Byte(CharLower(From));

                        for i:= ALength downto 1 do
                        begin
                                AC:= Byte(S[i]);
                                if (AC >= Byte('A'))and(AC <= Byte('Z')) then Inc(AC, 32);

                                if (AC = AF) then
                                begin
                                        S[i]:= From;
                                        Result:= True;

                                        Exit;
                                end;
                        end;
                 end;
                 //
                end;
        end;
end;

//##############################################################################
//##                           string support                                 ##
//##############################################################################

function StringUpper(const S: String): String;
var i, ALength: Integer;
begin
        Result:= S;
        ALength:= Length(Result);

        for i:= 1 to ALength do
        if (Byte(Result[i]) >= Byte('a'))and(Byte(Result[i]) <= Byte('z')) then Dec(Byte(Result[i]), 32);
end;


function StringLower(const S: String): String;
var i, ALength: Integer;
begin
        Result:= S;
        ALength:= Length(Result);

        for i:= 1 to ALength do
        if (Byte(Result[i]) >= Byte('A'))and(Byte(Result[i]) <= Byte('Z')) then Inc(Byte(Result[i]), 32);
end;


//#
function StringIn(const S: String; const A: array of String; CaseSensitive: Boolean = True): Boolean;
var i, ACount: Integer;
begin
        Result:= False;
        ACount:= High(A);

        if (ACount > 0) then
        begin
                for i:= 0 to ACount do
                if StringEqual(S, A[i], CaseSensitive) then
                begin
                        Result:= True;
                        Exit;
                end;
        end;
end;


//#
function StringEqual(const S1, S2: String; CaseSensitive: Boolean = True): Boolean;
var i, ALength: Integer;
    AC1, AC2: Byte;
begin
        Result:= Length(S1) = Length(S2);
        ALength:= Length(S1);

        if (Pointer(S1) <> Pointer(S2))and(Result) then
        begin
                case CaseSensitive of
                  True: Result:= (S1 = S2);
                  //
                  else begin
                        for i:= 1 to ALength do
                        begin
                                AC1:= Byte(S1[i]);
                                if (AC1 >= Byte('A'))and(AC1 <= Byte('Z')) then Inc(AC1, 32);

                                AC2:= Byte(S2[i]);
                                if (AC2 >= Byte('A'))and(AC2 <= Byte('Z')) then Inc(AC2, 32);

                                if (AC1 <> AC2) then
                                begin
                                        Result:= False;
                                        Exit;
                                end;
                         end;
                  end;
                  //
                end;
        end;
end;


function StringCompare(const S1, S2: String): Integer;
var i, ALength: Integer;
    AC1, AC2: Byte;
begin
        Result:= 0;
        ALength:= Min(Length(S1), Length(S2));

        if Pointer(S1) <> Pointer(S2) then
        begin
                for i:= 1 to ALength do
                begin
                        AC1:= Byte(S1[i]);
                        if (AC1 >= Byte('A'))and(AC1 <= Byte('Z')) then Inc(AC1, 32);

                        AC2:= Byte(S2[i]);
                        if (AC2 >= Byte('A'))and(AC2 <= Byte('Z')) then Inc(AC2, 32);

                        if (AC1 > AC2) then
                        begin
                                Result:= 1;
                                Exit;
                        end
                        else if (AC1 < AC2) then
                        begin
                                Result:= -1;
                                Exit;
                        end;
                end;

                if (Length(S1) > Length(S2)) then Result:= 1
                else if (Length(S1) < Length(S2)) then Result:= -1;
        end;
end;


//#
function StringIsCard(const S: String): Boolean;
var i, ALength: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                Result:= True;

                for i:= 1 to ALength do
                if (Byte(S[i]) < Byte('0'))or(Byte(S[i]) > Byte('9')) then
                begin
                        Result:= False;
                        Exit;
                end;
        end
        else Result:= False;
end;


function StringIsInt(const S: String): Boolean;
var i, ALength: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0)and((S[1] <> '-')or(ALength > 1)) then
        begin
                Result:= True;

                if (S[1] = '-') then i:= 2
                else i:= 1;

                for i:= i to ALength do
                if (Byte(S[i]) < Byte('0'))or(Byte(S[i]) > Byte('9')) then
                begin
                        Result:= False;
                        Exit;
                end;
        end
        else Result:= False;
end;


function StringIsHex(const S: String): Boolean;
var i, ALength: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                Result:= True;

                if (S[1] = '$') then i:= 2
                else i:= 1;

                for i:= i to ALength do
                if ((Byte(S[i]) < Byte('0'))or(Byte(S[i]) > Byte('9')))and
                   ((Byte(S[i]) < Byte('a'))or(Byte(S[i]) > Byte('f')))and
                   ((Byte(S[i]) < Byte('A'))or(Byte(S[i]) > Byte('F'))) then
                begin
                        Result:= False;
                        Exit;
                end;
        end
        else Result:= False;
end;


function StringIsExt(const S: String): Boolean;
var i, ALength: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0)and((S[1] <> '-')or(ALength > 1))and(StringMatchCount(S, '.') = 1) then
        begin
                Result:= True;

                if (S[1] = '-') then i:= 2
                else i:= 1;

                for i:= i to ALength do
                if ((Byte(S[i]) < Byte('0'))or(Byte(S[i]) > Byte('9')))and(S[i] <> '.') then
                begin
                        Result:= False;
                        Exit;
                end;
        end
        else Result:= False;
end;


function StringIsBool(const S: String): Boolean;
begin
        Result:= StringEqual(S, 'true', False)or(StringEqual(S, 'false', False));
end;


//#
function StringTrim(const S: String): String;
begin
        Result:= StringTrimLeft(StringTrimRight(S));
end;


function StringTrimLeft(const S: String): String;
var i, ALength: Integer;
begin
        i:= 1;
        ALength:= Length(S);
        while (i <= ALength)and(S[i] = ' ') do Inc(i);

        Result:= Copy(S, i, MAX_INT);
end;


function StringTrimRight(const S: String): String;
var i: Integer;
begin
        i:= Length(S);
        while (i > 0)and(S[i] = ' ') do Dec(i);

        Result:= Copy(S, 1, i);
end;


function StringTrimEllipse(const S: String; ALength: Integer) : String;
begin
        if (Length(S) <= ALength) then Result:= S
        else if (ALength < 3) then Result:= '...'
        else Result:= Copy(S, 1, ALength - 3) + '...';
end;


//#
function StringFill(const AFill: String; ALength: Integer): String;
var i, AFillLength: Integer;
begin
        AFillLength:= Length(AFill);

        if (ALength > 0)and(AFillLength > 0) then
        begin
                SetLength(Result, ALength);

                for i:= 1 to (ALength div AFillLength) do HeapCopyByte(Pointer(Integer(Result) + (i-1)*AFillLength), Pointer(AFill), AFillLength);
                HeapCopyByte(Pointer(Integer(Result) + (ALength div AFillLength)*AFillLength), Pointer(AFill), (ALength mod AFillLength));
        end;
end;


function StringFill(const AFill: Char; ALength: Integer): String;
begin
        if (ALength > 0) then
        begin
                SetLength(Result, ALength);

                HeapFillByte(Pointer(Result), Byte(AFill), ALength);
        end;
end;


//##############################################################################
//##                               char support                               ##
//##############################################################################

function CharUpper(C: Char): Char;
begin
        if (Byte(C) >= Byte('a'))and(Byte(C) <= Byte('z')) then Result:= Char(Byte(C) - 32)
        else Result:= C;
end;


function CharLower(C: Char): Char;
begin
        if (Byte(C) >= Byte('A'))and(Byte(C) <= Byte('Z')) then Result:= Char(Byte(C) + 32)
        else Result:= C;
end;


//#
function CharIn(C: Char; const A: array of Char; CaseSensitive: Boolean = True): Boolean;
var i, ACount: Integer;
begin
        Result:= False;
        ACount:= High(A);

        if (ACount > 0) then
        begin
                for i:= 0 to ACount do
                if CharEqual(C, A[i], CaseSensitive) then
                begin
                        Result:= True;
                        Exit;
                end;
        end;
end;


//#
function CharIsInt(C: Char): Boolean;
begin
        if (Byte(C) < Byte('0'))or(Byte(C) > Byte('9')) then Result:= True
        else Result:= False;
end;


function CharIsHex(C: Char): Boolean;
begin
        if ((Byte(C) < Byte('0'))or(Byte(C) > Byte('9')))and
           ((Byte(C) < Byte('a'))or(Byte(C) > Byte('f')))and
           ((Byte(C) < Byte('A'))or(Byte(C) > Byte('F'))) then Result:= True
        else Result:= False;
end;


//#
function CharEqual(C1, C2: Char; CaseSensitive: Boolean = True): Boolean;
begin
        case CaseSensitive of
          True: Result:= (C1 = C2);
          //
          else begin
                if (Byte(C1) >= Byte('A'))and(Byte(C1) <= Byte('Z')) then Inc(Byte(C1), 32);
                if (Byte(C2) >= Byte('A'))and(Byte(C2) <= Byte('Z')) then Inc(Byte(C2), 32);

                Result:= (C1 = C2);
          end;
          //
        end;
end;


function CharCompare(C1, C2: Char): Integer;
begin
        if (Byte(C1) >= Byte('A'))and(Byte(C1) <= Byte('Z')) then Inc(Byte(C1), 32);
        if (Byte(C2) >= Byte('A'))and(Byte(C2) <= Byte('Z')) then Inc(Byte(C2), 32);

        if (Byte(C1) > Byte(C2)) then Result:= 1
        else if (Byte(C1) < Byte(C2)) then Result:= -1
        else Result:= 0;
end;

//##############################################################################
//##                               pchar support                              ##
//##############################################################################

function PCharLen(const Str: PChar): Integer;
asm
        xchg    eax, edi
        xchg    edx, eax
        or      ecx, -1
        xor     eax, eax
        cmp     eax, edi
        je      @exit0
        repne   scasb
        dec     eax
        dec     eax
        sub     eax, ecx

@exit0:
        mov     edi, edx
end;


function PCharEnd(const Str: PChar): PChar;
asm
        mov     edx, edi
        mov     edi, eax
        mov     ecx, 0FFFFFFFFH
        xor     al, al
        repne   scasb
        lea     eax, [edi-1]
        mov     edi, edx
end;


function PCharMove(Dest: PChar; const Source: PChar; Count: Integer): PChar;
begin
        Result:= Dest;
        HeapMoveByte(Dest, Source, Count);
end;


function PCharCopy(Dest: PChar; const Source: PChar): PChar;
asm
        push    edi
        push    esi
        mov     esi, eax
        mov     edi, edx
        or      ecx, -1
        xor     al, al
        repne   scasb
        not     ecx
        mov     edi, esi
        mov     esi, edx
        mov     edx, ecx
        mov     eax, edi
        shr     ecx, 2
        rep     movsd
        mov     ecx, edx
        and     ecx, 3
        rep     movsb
        pop     esi
        pop     edi
end;


function PCharComp(const Str1, Str2: PChar): Integer;
asm
        push    edi
        push    esi
        mov     edi, edx
        xchg    esi, eax
        or      ecx, -1
        xor     eax, eax
        repne   scasb
        not     ecx
        mov     edi, edx
        xor     edx, edx
        repe    cmpsb
        mov     al, [esi-1]
        mov     dl, [edi-1]
        sub     eax, edx
        pop     esi
        pop     edi
end;


function PCharScan(const Str: PChar; Chr: Char): PChar;
asm
        push    edi
        push    eax
        mov     edi, str
        or      ecx, -1
        xor     al, al
        repne   scasb
        not     ecx
        pop     edi
        xchg    eax, edx
        repne   scasb

        xchg    eax, edi
        pop     edi

        je      @1
        xor     eax, eax
        ret

@1:    dec     eax
end;


function PCharRScan(const Str: PChar; Chr: Char): PChar;
asm
        push    edi
        mov     edi, str
        mov     ecx, 0FFFFFFFFH
        xor     al, al
        repne   scasb
        not     ecx
        std
        dec     edi
        mov     al, chr
        repne   scasb
        mov     eax, 0
        jne     @1
        mov     eax, edi
        inc     eax

@1:    cld
        pop     edi
end;


function PCharPos(const Str1, Str2: PChar): PChar;
asm
        push    edi
        push    esi
        push    ebx
        or      eax, eax
        je      @2
        or      edx, edx
        je      @2
        mov     ebx, eax
        mov     edi, edx
        xor     al, al
        mov     ecx, 0FFFFFFFFH
        repne   scasb
        not     ecx
        dec     ecx
        je      @2
        mov     esi, ecx
        mov     edi, ebx
        mov     ecx, 0FFFFFFFFH
        repne   scasb
        not     ecx
        sub     ecx, esi
        jbe     @2
        mov     edi, ebx
        lea     ebx, [esi-1]

@1:    mov     esi, edx
        lodsb
        repne   scasb
        jne     @2
        mov     eax, ecx
        push    edi
        mov     ecx, ebx
        repe    cmpsb
        pop     edi
        mov     ecx, eax
        jne     @1
        lea     eax, [edi-1]
        jmp     @3

@2:    xor     eax, eax

@3:    pop     ebx
        pop     esi
        pop     edi
end;


function PCharUpper(Str: PChar): PChar;
asm
        push    esi
        mov     esi, str
        mov     edx, str

@1:    lodsb
        or      al, al
        je      @2
        cmp     al, 'a'
        jb      @1
        cmp     al, 'z'
        ja      @1
        sub     al, 20h
        mov     [esi-1], al
        jmp     @1

@2:    xchg    eax, edx
        pop     esi
end;


function PCharLower(Str: PChar): PChar;
asm
        push    esi
        mov     esi, str
        mov     edx, str

@1:    lodsb
        or      al, al
        je      @2
        cmp     al, 'a'
        jb      @1
        cmp     al, 'z'
        ja      @1
        add     al, 20h
        mov     [esi-1], al
        jmp     @1

@2:    xchg    eax, edx
        pop     esi
end;

//##############################################################################
//##                            convert support                               ##
//##############################################################################

function CartToHex(Value: Cardinal): String;
begin
        if (Value <> 0) then
        begin
                Result:= '';

                while (Value <> 0) do
                begin
                        case (Value mod 16) of
                          0..9: Result:= Char(Byte('0') + (Value mod 16)) + Result;
                          10..15: Result:= Char(Byte('A') + (Value mod 16) - 10) + Result;
                        end;

                        Value:= Value div 16;
                end;
        end
        else Result:= '0';
end;


function IntToHex(Value: Integer): String;
begin
        if (Value <> 0) then
        begin
                Result:= '';
                Value:= Abs(Value);

                while (Value <> 0) do
                begin
                        case (Value mod 16) of
                          0..9: Result:= Char(Byte('0') + (Cardinal(Value) mod 16)) + Result;
                          10..15: Result:= Char(Byte('A') + (Cardinal(Value) mod 16) - 10) + Result;
                        end;

                        Cardinal(Value):= Cardinal(Value) div 16;
                end;
        end
        else Result:= '0';
end;


//#
function HexToCard(const S: String): Cardinal;
var ALength, i: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                Result:= 0;

                if (S[1] = '$') then i:= 2
                else i:= 1;

                for i:= i to ALength do
                begin
                        if (Byte(S[i]) >= Byte('0'))and(Byte(S[i]) <= Byte('9')) then
                                Result:= (Result shl 4) + Cardinal(Byte(S[i]) - Byte('0'))
                        else if (Byte(S[i]) >= Byte('A'))and(Byte(S[i]) <= Byte('F')) then
                                Result:= (Result shl 4) + Cardinal(Byte(S[i]) - Byte('A') + 10)
                        else if (Byte(S[i]) >= Byte('a'))and(Byte(S[i]) <= Byte('f')) then
                                Result:= (Result shl 4) + Cardinal(Byte(S[i]) - Byte('a') + 10)
                        else begin
                                Result:= 0;
                                Exit;
                        end;
                end;
        end
        else Result:= 0;
end;


function HexToInt(const S: String): Integer;
var ALength, i: Integer;
begin
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                Result:= 0;

                if (S[1] = '$') then i:= 2
                else i:= 1;

                for i:= i to ALength do
                begin
                        if (Byte(S[i]) >= Byte('0'))and(Byte(S[i]) <= Byte('9')) then
                                Result:= (Result shl 4) + (Byte(S[i]) - Byte('0'))
                        else if (Byte(S[i]) >= Byte('A'))and(Byte(S[i]) <= Byte('F')) then
                                Result:= (Result shl 4) + (Byte(S[i]) - Byte('A') + 10)
                        else if (Byte(S[i]) >= Byte('a'))and(Byte(S[i]) <= Byte('f')) then
                                Result:= (Result shl 4) + (Byte(S[i]) - Byte('a') + 10)
                        else begin
                                Result:= 0;
                                Exit;
                        end;
                end;
        end
        else Result:= 0;
end;


//#
function CardToStr(Value: Cardinal): String;
begin
        if (Value <> 0) then
        begin
                Result:= '';

                while (Value <> 0) do
                begin
                        Result:= Char(Byte('0') + (Value mod 10)) + Result;
                        Value:= Value div 10;
                end;
        end
        else Result:= '0';
end;


function IntToStr(Value: Integer): String;
var Lz: Boolean;
begin
        if (Value <> 0) then
        begin
                Result:= '';
                Lz:= (Value < 0);
                Value:= Abs(Value);

                while (Value <> 0) do
                begin
                        Result:= Char(Byte('0') + (Value mod 10)) + Result;
                        Value:= Value div 10;
                end;

                if Lz then Result:= '-' + Result;
        end
        else Result:= '0';
end;


function ExtToStr(Value: Extended; Decimals: Integer): String;
begin
        Result:= IntToStr(Trunc(Value)) + '.' + IntToStr(Trunc(Frac(Value)*Power(10, Decimals)));
end;


function BoolToStr(Value: Boolean): String;
begin
        case Value of
          True: Result:= 'true';
          else Result:= 'false';
        end;
end;


//#
function StrToCard(const S: String): Cardinal;
var ALength, i: Integer;
begin
        Result:= 0;
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                if (S[1] = '-') then i:= 2
                else i:= 1;

                for i:= i to ALength do
                begin
                        if (Byte(S[i]) >= Byte('0'))and(Byte(S[i]) <= Byte('9')) then
                                Result:= (Result * 10) + Cardinal(Byte(S[i]) - Byte('0'))
                        else begin
                                Result:= 0;
                                Exit;
                        end;
                end;
        end;
end;


function StrToInt(const S: String): Integer;
var ALength, i: Integer;
    Lz: Boolean;
begin
        Result:= 0;
        ALength:= Length(S);

        if (ALength > 0) then
        begin
                if (S[1] = '-') then
                begin
                        i:= 2;
                        Lz:= True;
                end
                else begin
                        i:= 1;
                        Lz:= False;
                end;

                for i:= i to ALength do
                begin
                        if (Byte(S[i]) >= Byte('0'))and(Byte(S[i]) <= Byte('9')) then
                                Result:= (Result * 10) + (Byte(S[i]) - Byte('0'))
                        else begin
                                Result:= 0;
                                Exit;
                        end;
                end;

                if Lz then Result:= -Result;
        end;
end;


function StrToExt(const S: String): Extended;
var i: Integer;
begin
        i:= StringMatchIndexLeft(S, '.');

        if (i <> 0) then Result:= StrToInt(Copy(S, 1, i - 1)) + StrToInt(Copy(S, i + 1, MAX_INT)) / Power(10, Length(S) - i)
        else Result:= StrToInt(S);
end;


function StrToBool(const S: String): Boolean;
begin
        if StringEqual(S, 'true', False) then Result:= True
        else Result:= False;
end;

//##############################################################################
//##                             integer support                              ##
//##############################################################################

procedure Restrict(var Nr: Integer; Min, Max: Integer);
begin
        if (Max < Min) then Max:= Min;

        if (Nr > Max) then Nr:= Max
        else if (Nr < Min) then Nr:= Min;
end;


procedure Restrict(var Nr: Cardinal; Min, Max: Cardinal);
begin
        if (Max < Min) then Max:= Min;

        if (Nr > Max) then Nr:= Max
        else if (Nr < Min) then Nr:= Min;
end;


procedure Restrict(var Nr: Extended; Min, Max: Extended);
begin
        if (Max < Min) then Max:= Min;

        if (Nr > Max) then Nr:= Max
        else if (Nr < Min) then Nr:= Min;
end;


procedure Restrict(var APoint: TPoint; ABounds: TRect);
begin
        if (ABounds.Right < ABounds.Left) then ABounds.Right:= ABounds.Left;

        if (APoint.X > ABounds.Right) then APoint.X:= ABounds.Right
        else if (APoint.X < ABounds.Left) then APoint.X:= ABounds.Left;

        if (ABounds.Bottom < ABounds.Top) then ABounds.Bottom:= ABounds.Top;

        if (APoint.Y > ABounds.Bottom) then APoint.Y:= ABounds.Bottom
        else if (APoint.Y < ABounds.Top) then APoint.Y:= ABounds.Top;
end;


procedure Restrict(var ARect: TRect; ABounds: TRect);
begin
        if (ARect.Left >= ABounds.Right) then
        begin
                ARect.Left:= ARect.Right;
                Exit;
        end
        else if (ARect.Left < ABounds.Left) then ARect.Left:= ABounds.Left;

        if (ARect.Right <= ABounds.Left) then
        begin
                ARect.Right:= ARect.Left;
                Exit;
        end
        else if (ARect.Right > ABounds.Right) then ARect.Right:= ABounds.Right;


        if (ARect.Top >= ABounds.Bottom) then
        begin
                ARect.Top:= ARect.Bottom;
                Exit;
        end
        else if (ARect.Top < ABounds.Top) then ARect.Top:= ABounds.Top;

        if (ARect.Bottom <= ABounds.Top) then
        begin
                ARect.Bottom:= ARect.Top;
                Exit;
        end
        else if (ARect.Bottom > ABounds.Bottom) then ARect.Bottom:= ABounds.Bottom;
end;


//#
function Ceil(X: Extended): Integer;
begin
        Result:= Trunc(X);
        if (Frac(X) > 0) then Inc(Result);
end;


function Floor(X: Extended): Integer;
begin
        Result:= Trunc(X);
        if (Frac(X) < 0) then Dec(Result);
end;


//#
function Min(X, Y: Integer): Integer;
begin
        if (X >= Y) then Result:= Y
        else Result:= X;
end;


function Min(X, Y: Cardinal): Cardinal;
begin
        if (X >= Y) then Result:= Y
        else Result:= X;
end;


function Min(X, Y: Extended): Extended;
begin
        if (X >= Y) then Result:= Y
        else Result:= X;
end;


//#
function Max(X, Y: Integer): Integer;
begin
        if (X >= Y) then Result:= X
        else Result:= Y;
end;


function Max(X, Y: Cardinal): Cardinal;
begin
        if (X >= Y) then Result:= X
        else Result:= Y;
end;


function Max(X, Y: Extended): Extended;
begin
        if (X >= Y) then Result:= X
        else Result:= Y;
end;


//#
procedure Exchange(var X, Y: Integer);
var i: Integer;
begin
        i:= X;
        X:= Y;
        Y:= i;
end;


procedure Exchange(var X, Y: Cardinal);
var i: Cardinal;
begin
        i:= X;
        X:= Y;
        Y:= i;
end;


procedure Exchange(var X, Y: Extended);
var i: Extended;
begin
        i:= X;
        X:= Y;
        Y:= i;
end;


//#
function Abs(X: Integer): Integer;
begin
        if (X < 0) then Result:= -X
        else Result:= X;
end;


function Abs(X: Extended): Extended;
begin
        if (X < 0) then Result:= -X
        else Result:= X;
end;


//#
function Sgn(X: Integer): Integer;
begin
        if (X < 0) then Result:= -1
        else if (X > 0) then Result:= 1
        else Result:= 0;
end;


function Sgn(X: Extended): Integer;
begin
        if (X < 0) then Result:= -1
        else if (X > 0) then Result:= 1
        else Result:= 0;
end;


//#
function Digits(Nr: Integer): Integer;
begin
        Result:= 0;

        while (Nr > 0) do
        begin
                Nr:= Nr div 10;
                Inc(Result);
        end;
end;


function Digits(Nr: Cardinal): Integer;
begin
        Result:= 0;

        while (Nr > 0) do
        begin
                Nr:= Nr div 10;
                Inc(Result);
        end;
end;


//#
function Power(Base: Extended; Exponent: Integer): Extended;
asm
        mov     ecx, eax
        cdq
        fld1                      { Result := 1 }
        xor     eax, edx
        sub     eax, edx          { eax := Abs(Exponent) }
        jz      @@3
        fld     Base
        jmp     @@2
@@1:    fmul    ST, ST            { X := Base * Base }
@@2:    shr     eax, 1
        jnc     @@1
        fmul    ST(1), ST         { Result := Result * X }
        jnz     @@1
        fstp    st                { pop X from FPU stack }
        cmp     ecx, 0
        jge     @@3
        fld1
        fdivrp                    { Result := 1 / Result }
@@3:
        fwait
end;


//#
function LowHigh(ALow, AHigh: SmallInt): Integer;
begin
        Result:= (ALow and $0000FFFF)or(AHigh shl 16);
end;


function LowInt(I: Integer): SmallInt;
begin
        Result:= SmallInt(I and $0000FFFF);
end;


function HighInt(I: Integer): SmallInt;
begin
        Result:= SmallInt(I shr 16);
end;

//##############################################################################
//##                            path manipulation                             ##
//##############################################################################

function PathAddSlash(const APath: String): String;
var ALength: Integer;
begin
        ALength:= Length(APath);

	if (ALength > 0)and(APath[ALength] <> '\') then Result:= APath + '\'
	else Result:= APath;
end;


function PathRemoveSlash(const APath: String): String;
var ALength: Integer;
begin
        ALength:= Length(APath);

	if (ALength > 0)and(APath[ALength] = '\') then Result:= Copy(APath, 1, Length(APath) - 1)
	else Result:= APath;
end;


function PathExtractFileName(const APath: String): String;
var i: Integer;
begin
        i:= StringMatchIndexRight(APath, '\');

        Result:= Copy(APath, i + 1, MAX_INT);
end;


function PathExtractDrive(const APath: String): Char;
begin
        if (Length(APath) >= 2)and(APath[2] = ':') then Result:= APath[1]
        else Result:= #0;
end;


function PathExtract(const APath: String): String;
var i: Integer;
begin
        i:= StringMatchIndexRight(APath, '\');

        Result:= Copy(APath, 1, i);
end;


function PathDeph(const APath: String): Integer;
begin
        Result:= StringMatchCount(PathRemoveSlash(APath), '\');
end;


function PathShort(const APath: String): String;
var Buffer: array[0..MAX_PATH - 1] of Char;
begin
        SetString(Result, Buffer, GetShortPathName(PChar(APath), Buffer, SizeOf(Buffer)));
end;


function PathLong(const APath: String): String;
var Buffer: array[0..MAX_PATH - 1] of Char;
begin
        SetString(Result, Buffer, GetLongPathName(PChar(APath), Buffer, SizeOf(Buffer)));
end;


function PathComplete(const APath: String): String;
begin
        if (Length(APath) >= 2)and(APath[2] = ':') then Result:= PathAddSlash(DirGetCurrent) + APath
        else Result:= APath;
end;

//##############################################################################
//##                           file manipulation                              ##
//##############################################################################

function FileExtractName(const AFileName: String): String;
var i, j: Integer;
begin
        j:= StringMatchIndexRight(AFileName, '\');
        i:= StringMatchIndexRight(AFileName, '.');

        if (i <> 0) then Result:= Copy(AFileName, j + 1, i - j - 1)
        else Result:= Copy(AFileName, j + 1, MAX_INT);
end;


function FileChangeName(const AFileName, AName: String): String;
var i, j: Integer;
begin
        j:= StringMatchIndexLeft(AFilename, '\');
        i:= StringMatchIndexLeft(AFilename, '.');

        if (i <> 0) then Result:= Copy(AFileName, 1, j) + AName + Copy(AFileName, i, MAX_INT)
        else Result:= Copy(AFileName, 1, j) + AName;
end;


function FileExtractExt(const AFileName: String): String;
var i: Integer;
begin
        i:= StringMatchIndexRight(AFileName, '.');

        if (i <> 0) then Result:= Copy(AFileName, i + 1, MAX_INT)
        else Result:= '';
end;


function FileChangeExt(const AFileName, AExtension: String): String;
var i: Integer;
begin
        i:= StringMatchIndexLeft(AFilename, '.');

        if (i <> 0) then Result:= Copy(AFileName, 1, i) + AExtension
        else Result:= AFileName + AExtension;
end;


function FileExists(const AFileName: String): Boolean;
var Code: Integer;
begin
        Code:= HWindows.GetFileAttributes(PChar(AFileName));

        // check if the function failed and that the
        // filename is not a directory.
        Result:= (Code <> -1)and(FA_DIRECTORY and Code = 0);
end;


function FileSize(const AFileName: String): Int64;
var FindData: TSearchRec;
begin
        if FindFirst(AFileName, 0, FindData) then
        begin
                Result:= FindData.Size;
                FindClose(FindData);
        end
        else Result:= 0;
end;


function FileGetAttr(const AFileName: String): Integer;
begin
        Result:= HWindows.GetFileAttributes(PChar(AFileName));
        if (Result = -1) then Result:= 0;
end;


function FileSetAttr(const AFileName: String; AAttr: Integer): Boolean;
begin
        Result:= HWindows.SetFileAttributes(PChar(AFileName), AAttr);
end;


//#
function FileOpen(const AFileName: String; AOpenMode, AShareMode: Cardinal; AAcessMode: Cardinal = FM_ACCESS_UNKNOWN; AFileAttr: Cardinal = FA_NORMAL): Integer;
begin
        Result:= HWindows.CreateFile(PChar(AFileName), AOpenMode, AShareMode, nil, OPEN_EXISTING, AFileAttr or AAcessMode, 0);
        if (Result = -1) then Result:= 0;
end;


function FileCreate(const AFileName: String; AOpenMode, AShareMode: Cardinal; AAcessMode: Cardinal = FM_ACCESS_UNKNOWN; AFileAttr: Cardinal = FA_NORMAL): Integer;
begin
        Result:= HWindows.CreateFile(PChar(AFileName), AOpenMode, AShareMode, nil, CREATE_ALWAYS, AFileAttr or AAcessMode, 0);
        if (Result = -1) then Result:= 0;
end;


function FileRead(AHandle: Cardinal; var ABuffer; ACount: Integer): Integer;
begin
        if (not HWindows.ReadFile(AHandle, ABuffer, ACount, Cardinal(Result), nil)) then Result:= 0;
end;


function FileWrite(AHandle: Cardinal; const ABuffer; ACount: Integer): Integer;
begin
        if (not HWindows.WriteFile(AHandle, ABuffer, ACount, Cardinal(Result), nil)) then Result:= 0;
end;


function FileSeek(AHandle: Cardinal; AOffset: Int64; AOrigin: Integer): Int64;
begin
        Result:= AOffset;

        Int64Rec(Result).Cardinals[0]:= HWindows.SetFilePointer(AHandle, Int64Rec(Result).Cardinals[0],
                                                                @Int64Rec(Result).Cardinals[1], AOrigin);
end;


procedure FileClose(AHandle: Cardinal);
begin
        HWindows.CloseHandle(AHandle);
end;


//#
function FileCopy(const AName, ADestinationName: String; Overwrite: Boolean = False): Boolean;
begin
        Result:= HWindows.CopyFile(PChar(AName), PChar(ADestinationName), not Overwrite);
end;


function FileDelete(const AFileName: String): Boolean;
begin
        Result:= HWindows.DeleteFile(PChar(AFileName));
end;


function FileRename(const AOldName, ANewName: String; Overwrite: Boolean = False): Boolean;
begin
        case (Overwrite and FileExists(ANewName)) of
         True:
         begin
                Result:= HWindows.CopyFile(PChar(AOldName), PChar(ANewName), False);
                if Result then HWindows.DeleteFile(PChar(AOldName));
         end;
         //
         else Result:= HWindows.MoveFile(PChar(AOldName), PChar(ANewName));
         //
        end;
end;


//#
function DirCreate(const ADir: String): Boolean;
begin
        Result:= HWindows.CreateDirectory(PChar(ADir), nil);
end;


function DirCopy(const ADir, ADestinationDir: String; Overwrite: Boolean = False): Boolean;
begin
        case DirEmpty(ADir) of
         True:
         begin
                Result:= HWindows.CreateDirectory(PChar(ADir), nil);
                if Result then HWindows.RemoveDirectory(PChar(ADir));
         end;
         //
         else begin
                Result:= False;
         end;
         //
        end;
end;


function DirDelete(const ADir: String): Boolean;
begin
        case DirEmpty(ADir) of
         True: Result:= HWindows.RemoveDirectory(PChar(ADir));
         //
         else begin
                Result:= False;
         end;
         //
        end;
end;


function DirRename(const AOldDir, ANewDir: String; Overwrite: Boolean = False): Boolean;
begin
        case (Overwrite and DirExists(ANewDir)) of
         True:
         begin
                Result:= DirCopy(AOldDir, ANewDir, Overwrite);
                if Result then DirDelete(AOldDir);
         end;
         //
         else Result:= HWindows.MoveFile(PChar(AOldDir), PChar(ANewDir));
         //
        end;
end;


function DirExists(const ADir: String): Boolean;
var Code: Integer;
begin
        Code:= HWindows.GetFileAttributes(PChar(ADir));

        // check if the function failed and that the
        // filename is a directory.
        Result:= (Code <> -1)and(FA_DIRECTORY and Code <> 0);
end;


function DirEmpty(const ADir: String): Boolean;
var FindData: TSearchRec;
begin
        if FindFirst(PathAddSlash(ADir) + '*.*', 0, FindData) then
        begin
                Result:= False;
                FindClose(FindData);
        end
        else Result:= True;
end;


function DirGetCurrent: String;
var Buffer: array[0..MAX_PATH - 1] of Char;
begin
        SetString(Result, Buffer, HWindows.GetCurrentDirectory(SizeOf(Buffer), Buffer));
end;


function DirSetCurrent(const ADir: String): Boolean;
begin
        Result:= HWindows.SetCurrentDirectory(PChar(ADir));
end;

//##############################################################################
//##                        file search functions                             ##
//##############################################################################

function FindFirst(const APath: String; AExcludeAttr: Integer; var F: TSearchRec): Boolean;
begin
        with F do
        begin
                ExcludeAttr:= AExcludeAttr or FA_VOLUMEID;
                FindHandle:= HWindows.FindFirstFile(PChar(APath), FindData);

                if (FindHandle <> INVALID_HANDLE_VALUE) then
                begin
                        Result:= True;

                        HWindows.FileTimeToSystemTime(FindData.ftCreationTime, CreateTime);
                        HWindows.FileTimeToSystemTime(FindData.ftLastAccessTime, AccessTime);
                        HWindows.FileTimeToSystemTime(FindData.ftLastWriteTime, WriteTime);
                        Int64Rec(Size).Cardinals[0]:= FindData.nFileSizeLow;
                        Int64Rec(Size).Cardinals[1]:= FindData.nFileSizeHigh;
                        Attr:= FindData.dwFileAttributes;
                        Name:= FindData.cFileName;
                end
                else Result:= False;
        end;
end;


function FindNext(var F: TSearchRec): Boolean;
begin
        with F do
        begin
                Result:= False;

                while ((FindData.dwFileAttributes and ExcludeAttr <> 0)and(ExcludeAttr <> 0))or
                      (StringIn(FindData.cFileName, ['.', '..']))or(not Result) do
                begin
                        Result:= HWindows.FindNextFile(FindHandle, FindData);
                        if (not Result) then Exit;
                end;

                HWindows.FileTimeToSystemTime(FindData.ftCreationTime, CreateTime);
                HWindows.FileTimeToSystemTime(FindData.ftLastAccessTime, AccessTime);
                HWindows.FileTimeToSystemTime(FindData.ftLastWriteTime, WriteTime);
                Int64Rec(Size).Cardinals[0]:= FindData.nFileSizeLow;
                Int64Rec(Size).Cardinals[1]:= FindData.nFileSizeHigh;
                Attr:= FindData.dwFileAttributes;
                Name:= FindData.cFileName;
        end;
end;


procedure FindClose(var F: TSearchRec);
begin
        with F do
        begin
                if (FindHandle <> INVALID_HANDLE_VALUE) then
                begin
                        HWindows.FindClose(FindHandle);

                        FindHandle:= INVALID_HANDLE_VALUE;
                end;
        end;
end;

//##############################################################################
//##                             drive functions                              ##
//##############################################################################

function DriveGetName(ADrive: Char): String;
var Buffer: array [0..MAX_PATH - 1] of Char;
    ErrorMode: Integer;
    SystemFlags: Cardinal;
begin
	ErrorMode:= HWindows.SetErrorMode(SEM_FAILCRITICALERRORS or SEM_NOOPENFILEERRORBOX);

        if GetVolumeInformation(PChar(ADrive + ':\'), Buffer, SizeOf(Buffer), nil, SystemFlags, SystemFlags, nil, 0) then
                SetString(Result, Buffer, PCharLen(Buffer))
        else Result:= '';

        SetErrorMode(ErrorMode);
end;


procedure DriveSetName(ADrive: Char; const AName: String);
begin
        SetVolumeLabel(PChar(ADrive + ':\'), PChar(AName));
end;


function DriveFree(ADrive: Char): Int64;
var TotalSpace: Int64;
begin
        if (not GetDiskFreeSpaceEx(PChar(ADrive + ':\'), Result, TotalSpace, nil)) then Result:= 0;
end;


function DriveSize(ADrive: Char): Int64;
var FreeSpace: Int64;
begin
        if (not GetDiskFreeSpaceEx(PChar(ADrive + ':\'), FreeSpace, Result, nil)) then Result:= 0;
end;


function DriveID(ADrive: Char): String;
var ErrorMode: Integer;
    SystemFlags: Cardinal;
    Serial: Cardinal;
begin
	ErrorMode:= HWindows.SetErrorMode(SEM_FAILCRITICALERRORS or SEM_NOOPENFILEERRORBOX);

        if GetVolumeInformation(PChar(ADrive + ':\'), nil, 0, @Serial, SystemFlags, SystemFlags, nil, 0) then
                Result:= '$' + IntToHex(Serial)
        else Result:= '';

	SetErrorMode(ErrorMode);
end;


function DriveSystem(ADrive: Char): String;
var Buffer: array[0..12] of Char;
    SystemFlags: Cardinal;
begin
        if HWindows.GetVolumeInformation(PChar(ADrive + ':\'), nil, 0, nil, SystemFlags, SystemFlags, Buffer, 12) then
                SetString(Result, Buffer, PCharLen(Buffer))
        else Result:= '';
end;


function DriveType(ADrive: Char): Integer;
begin
        DriveType:= HWindows.GetDriveType(PChar(ADrive + ':\'))
end;


function DriveCount: Integer;
var DriveBits: Set of 0..25;
    i: Integer;
begin
        Integer(DriveBits):= HWindows.GetLogicalDrives;

        Result:= 0;
        for i:= 0 to 25 do
                if (i in DriveBits) then Inc(Result);
end;


function DriveString: String;
var DriveBits: Set of 0..25;
    i: Integer;
begin
        Integer(DriveBits):= HWindows.GetLogicalDrives;

        for i:= 0 to 25 do
                if (i in DriveBits) then Result:= Result + Char(Byte(FIRST_DRIVE) + i);
end;


function DriveReady(ADrive: Char): Boolean;
var ErrorMode: Integer;
begin
	ErrorMode:= HWindows.SetErrorMode(SEM_FAILCRITICALERRORS or SEM_NOOPENFILEERRORBOX);

        // try to access the drive, it doesn't really matter how we access the
        // drive and as such calling disksize is more or less a random choice.
        // the call to seterrormode supresses the system provided error dialog
        // if there is no disk in the drive and causes the to disksize to fail.
        Result:= (DriveSize(ADrive) <> -1);

	SetErrorMode(ErrorMode);
end;


function DrivesBitmaskToString(ABitmask: Integer): String;
var i: Integer;
    DriveBits: Set of 0..25;
begin
        Integer(DriveBits):= ABitmask;

        for i:= 0 to 25 do
        if (i in DriveBits) then
                Result:= Result + Char(Byte(FIRST_DRIVE) + i);
end;

//##############################################################################
//##                            network functions                             ##
//##############################################################################

function NetworkAddDrive(const Resource: String; ADrive: Char; const Password: String = ''): Boolean;
begin
        Result:= HNetwork.WNetAddConnection(PChar(Resource), PChar(Password), PChar(String(ADrive + ':'))) = NO_ERROR;
end;


function NetworkRemoveDrive(ADrive: Char; Force: Boolean = False): Boolean;
begin
        Result:= HNetwork.WNetCancelConnection2(PChar(String(ADrive + ':')), 0, Force) = NO_ERROR;
end;


function NetworkUniversalName(ADrive: Char): String;
var Buffer: PUniversalNameInfo;
    Size: Cardinal;
begin
        Size:= MAX_PATH;
        GetMem(Buffer, Size);

        HNetwork.WNetGetUniversalName(PChar(String(ADrive + ':\')), UNIVERSAL_NAME_INFO_LEVEL, Buffer, Size);

        SetString(Result, Buffer.lpUniversalName, PCharLen(Buffer.lpUniversalName));
        FreeMem(Buffer);
end;


function NetworkComputerName: String;
var Buffer: array[0..MAX_COMPUTERNAME_LENGTH - 1] of Char;
    Size: Cardinal;
begin
        Size:= SizeOf(Buffer);
        HWindows.GetComputerName(Buffer, Size);
        SetString(Result, Buffer, Size - 1);
end;


function NetworkUserName: String;
var Buffer: array[0..MAX_PATH - 1] of Char;
    Size: Cardinal;
begin
        Size:= SizeOf(Buffer);
        HWindows.GetUserName(Buffer, Size);
        SetString(Result, Buffer, Size - 1);
end;

//##############################################################################
//##                        default thread function                           ##
//##############################################################################

function ThreadProc(AThread: TAThread): Integer;
begin
        try
                try
                  if (not AThread.FTerminated) then AThread.Execute;
                except
                  (*++++++++++++++*)
                  (**) HandleException('ThreadProc');
                  (*++++++++++++++*)
                end;
        finally
                  (*==============*)
                  (**) if Assigned(AThread.FOnTerminate) then AThread.FOnTerminate(AThread);
                  (*==============*)

                  Result:= AThread.GetExitCode;
                  AThread.FTerminated:= True;
                  AThread.FFinished:= True;


                  (*><><><><><><><*)
                  (**) ThreadManager.SetExclusive(FreeOnTerminateLock);
                  (*><><><><><><><*)

                    if AThread.FFreeOnTerminate then FreeNil(AThread);

                  (*><><><><><><><*)
                  (**) ThreadManager.ClearExclusive(FreeOnTerminateLock);
                  (*><><><><><><><*)


                  // this needs to be called last because windows will assume the
                  // control of the thread from this point.
                  HWindows.ExitThread(Result);
        end;
end;

//#############################################################################
//##                        system functions                                 ##
//#############################################################################

function CpuSpeed: Extended;
var TimerHi, TimerLo: Cardinal;
begin
          asm
              dw 310FH
              mov TimerLo, eax
              mov TimerHi, edx
          end;

          HWindows.Sleep(1000);

          asm
              dw 310FH
              sub eax, TimerLo
              sbb edx, TimerHi
              mov TimerLo, eax
              mov TimerHi, edx
          end;

          Result:= TimerLo / (1000 * 1000);
end;


function CpuMmx: Boolean;
asm
        push    ebx
        xor     eax, eax     // result := false
        pushfd               // determine if the processor supports the cpuid command.
        pop     edx
        mov     ecx, edx
        xor     edx, $200000
        push    edx
        popfd
        pushfd
        pop     edx
        xor     ecx, edx
        jz      @1           // no cpuid support so we can't even get to the feature information.
        push    edx
        popfd

        mov     eax, 1
        dw      $a20f        // cpuid, eax contains now version info and edx feature information.
        mov     ebx, eax     // free eax to get the result value.
        xor     eax, eax     // result:= false
        cmp     ebx, $50
        jb      @1           // if processor family is < 5 then it is not a pentium class processor.
        test    edx, $800000
        jz      @1           // if the mmx bit is not set then we don't have mmx.
        inc     eax          // result:= true
@1:
        pop     ebx
end;


procedure CpuEmms;
asm
        db      $0f, $77
end;


function ProcedureToMethod(Self: TObject; ProcAddr: pointer): TMethod;
begin
        with Result do
        begin
                Data:= Self;
                Code:= ProcAddr;
        end;
end;


function MethodToProcedure(Self: TObject; MethodAddr: Pointer): Pointer;
var Mtp: PMethodToProc absolute Result;
begin
        GetMem(Mtp, SizeOf(TMethodToProc));

        with Mtp^ do
        begin
                PopEax:= $58;
                PushSelf.OpCode:= $68;
                PushSelf.Self:= self;
                PushEax:= $50;
                Jump.OpCode:= $FF;
                Jump.ModRm:= $25;
                Jump.PTarget:= @Jump.Target;
                Jump.Target:= MethodAddr;
        end;
end;


function MethodToProcedure(Method: TMethod): Pointer;
begin
        Result:= MethodToProcedure(TObject(Method.Data), Method.Code);
end;

//##############################################################################
//##                             windows remote                               ##
//##############################################################################

procedure WindowsShutDown(Force: Boolean = False);
begin
        case Force of
          True: HWindows.ExitWindowsEx(EWX_POWEROFF or EWX_FORCE or EWX_SHUTDOWN, 0);
          else HWindows.ExitWindowsEx(EWX_POWEROFF or EWX_SHUTDOWN, 0);
        end;
end;


procedure WindowsReboot(Force: Boolean = False);
begin
        case Force of
          True: HWindows.ExitWindowsEx(EWX_REBOOT or EWX_FORCE, 0);
          else HWindows.ExitWindowsEx(EWX_REBOOT, 0);
        end;
end;


function WindowsVersion: Integer;
var VersionInfo: TOSVersionInfo;
begin
        VersionInfo.dwOSVersionInfoSize:= SizeOf(TOSVersionInfo);
        HWindows.GetVersionEx(VersionInfo);

        with VersionInfo do
        case dwPlatformId of
          VER_PLATFORM_WIN32s: Result:= WINDOWS_31;
          //
          VER_PLATFORM_WIN32_WINDOWS:
          begin
                  Result:= WINDOWS_95;

                  if (dwMajorVersion = 4) then
                  begin
                       if (dwMinorVersion >= 90) then Result:= WINDOWS_ME
                       else if (dwMinorVersion >= 10) then
                       begin
                            if (dwBuildNumber >= 2222) then Result:= WINDOWS_98SE
                            else if (dwBuildNumber >= 1998) then Result:= WINDOWS_98;
                       end;
                  end;
          end;
          //
          VER_PLATFORM_WIN32_NT:
          begin
                 Result:= WINDOWS_NT;

                 if (dwMajorVersion = 5) then
                 begin
                         if (dwMinorVersion >= 1) then Result:= WINDOWS_XP
                         else Result:= WINDOWS_2000;
                 end;
          end;
          //
          else Result:= 0;
          //
        end;
end;


function WindowsFamily: Integer;
var VersionInfo: TOSVersionInfo;
begin
        VersionInfo.dwOSVersionInfoSize:= SizeOf(TOSVersionInfo);
        HWindows.GetVersionEx(VersionInfo);

        case VersionInfo.dwPlatformId of
          VER_PLATFORM_WIN32s: Result:= WINFAMILY_3X;
          VER_PLATFORM_WIN32_WINDOWS: Result:= WINFAMILY_9X;
          VER_PLATFORM_WIN32_NT: Result:= WINFAMILY_NT;
          //
          else Result:= 0;
        end;
end;


function WindowsMemory(MemRequest: Integer): Cardinal;
var MS: TMemoryStatus;
begin
        MS.dwLength:= SizeOf(TMemoryStatus);
        HWindows.GlobalMemoryStatus(MS);

        case MemRequest of
         MEMORY_PERCENT: Result:= MS.dwMemoryLoad;
         MEMORY_SIZE: Result:= MS.dwTotalPhys;
         MEMORY_FREE: Result:= MS.dwAvailPhys;
         MEMORY_SWAPSIZE: Result:= MS.dwTotalPageFile;
         MEMORY_SWAPFREE: Result:= MS.dwAvailPageFile;
         //
         else Result:= 0;
        end;
end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#########################                            #########################
//#########################                            #########################
//##                         STREAMS, FILE PATH DRIVE                         ##
//##                          MANIPULATION FUNCTIONS                          ##
//#########################                            #########################
//#########################                            #########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                               FILESTREAM                                 ##
//##############################################################################


{ TAFileStream }




constructor TAFileStream.Create(const AFileName: String; AOpenMode, AShareMode: Cardinal; AAcessMode: Cardinal = FM_ACCESS_UNKNOWN; AFileAttr: Cardinal = FA_NORMAL);
begin
FFileName:= AFileName;



        FHandle:= 0;



        if FileExists(FFileName) then FHandle:= FileOpen(FileName, AOpenMode and (not FM_EMPTY), AShareMode, AAcessMode, AFileAttr)
        else FHandle:= FileCreate(FileName, AOpenMode and (not FM_EMPTY), AShareMode, AAcessMode, AFileAttr);

        (*++++++++++++++*)
        (**) if (FHandle = 0) then RaiseException('Cannot create stream for file "' + FFileName + '".', 'TAFileStream.Create', Self, True);
        (*++++++++++++++*)

        if (AOpenMode and FM_EMPTY <> 0) then SetSize(0);
end;


destructor TAFileStream.Destroy;
begin
        FlushCache;

        if (FHandle <> 0) then FileClose(FHandle);



inherited;
end;


function TAFileStream.GetPosition: Int64;
begin
        Result:= FileSeek(FHandle, 0, FM_SEEK_CURRENT);
end;


function TAFileStream.GetSize: Int64;
var Pos: Integer;
begin
        // get current position.
        Pos:= FileSeek(FHandle, 0, FM_SEEK_CURRENT);

        // seek to the end of file.
        Result:= FileSeek(FHandle, 0, FM_SEEK_END);

        // get back to the stored position.
        FileSeek(FHandle, Pos, FM_SEEK_BEGIN);
end;


procedure TAFileStream.SetPosition(APosition: Int64);
begin
        FileSeek(FHandle, APosition, FM_SEEK_BEGIN);
end;


procedure TAFileStream.SetSize(ASize: Int64);
begin
        FileSeek(FHandle, ASize, FM_SEEK_BEGIN);

        (*++++++++++++++*)
        (**) if (not HWindows.SetEndOfFile(FHandle)) then RaiseException('Cannot setsize stream for file "' + FFileName + '".', 'TAFileStream.SetSize', Self);
        (*++++++++++++++*)
end;


function TAFileStream.Read(var ABuffer; ACount: Integer): Integer;
begin
        Result:= FileRead(FHandle, ABuffer, ACount);
end;


function TAFileStream.Write(const ABuffer; ACount: Integer): Integer;
begin
        Result:= FileWrite(FHandle, ABuffer, ACount);
end;


procedure TAFileStream.ReadBuffer(var ABuffer; ACount: Integer);
var Read: Integer;
begin
        Read:= FileRead(FHandle, ABuffer, ACount);

        if (ACount <> 0)and(Read <> ACount) then
                (*++++++++++++++*)
                (**) RaiseException('Cannot readbuffer from stream for file "' + FFileName + '". Requested ' + IntToStr(ACount) + ', read ' + IntToStr(Read) + '.', 'TAFileStream.ReadBuffer', Self);
                (*++++++++++++++*)
end;


procedure TAFileStream.WriteBuffer(const ABuffer; ACount: Integer);
var Write: Integer;
begin
        Write:= FileWrite(FHandle, ABuffer, ACount);

        if (ACount <> 0)and(Write <> ACount) then
                (*++++++++++++++*)
                (**) RaiseException('Cannot writebuffer to stream for file "' + FFileName + '". Requested ' + IntToStr(ACount) + ', written ' + IntToStr(Write) + '.', 'TAFileStream.WriteBuffer', Self);
                (*++++++++++++++*)
end;


function TAFileStream.Seek(AOffset: Int64; AOrigin: Integer): Int64;
begin
        Result:= FileSeek(FHandle, AOffset, AOrigin);
end;


procedure TAFileStream.FlushCache;
begin
        HWindows.FlushFileBuffers(FHandle);
end;


function TAFileStream.BOF: Boolean;
begin
        Result:= (GetPosition = 0);
end;


function TAFileStream.EOF: Boolean;
begin
        Result:= (GetPosition = GetSize);
end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##################                                        ####################
//##################                                        ####################
//##                  POINTER LIST, STRING LIST, RECT LIST                    ##
//##                    THREAD, THREADLIST, THREADMANAGER                     ##
//##################                                        ####################
//##################                                        ####################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                                  LIST                                    ##
//##############################################################################


{ TAList }




constructor TAList.Create(AAllowLock: Boolean = False);
begin
{$IFDEF LIST_THREADSAFE}
 if (FAllowLock) then FLock:= TALock.Create('List lock');
{$ENDIF}

FAllowLock:= AAllowLock;



        FCount:= 0;
        FCapacity:= 0;
        FDelta:= 4;



        SetCapacity(FDelta);
end;


destructor TAList.Destroy;
begin
        SetCapacity(0);



{$IFDEF LIST_THREADSAFE}
 if (FAllowLock) then FreeNil(FLock);
{$ENDIF}

inherited;
end;


function TAList.GetItem(AIndex: Integer): Pointer;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then Result:= FList^[AIndex]
        else begin
                Result:= nil;

                {$IFDEF LIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Getitem index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TAList.GetItem', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.SetItem(AIndex: Integer; AItem: Pointer);
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then FList^[AIndex]:= AItem
        else begin
                {$IFDEF LIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Setitem index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TAList.SetItem', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.SetCapacity(ACapacity: Integer);
begin
        if (ACapacity <> FCapacity) then
        begin
                ReallocMem(FList, ACapacity * SizeOf(Pointer));
                FCapacity:= ACapacity;
        end;
end;


procedure TAList.Grow;
begin
        // set the list gradularity according
        // to how many items are sored in.

        if (FCapacity > 64) then FDelta:= 32
        else if (FCapacity > 32) then FDelta:= 16
        else FDelta:= 4;

        SetCapacity(FCapacity + FDelta);
end;


function TAList.Add(AItem: Pointer): Integer;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= FCount;

        // if the list is full then reallocate memory for it.
        if (Result = FCapacity) then Grow;

        FList^[Result]:= AItem;
        Inc(FCount);

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.Delete(AIndex: Integer);
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then
        begin
                Dec(FCount);

                // if the index is between other items then move the items next to it.
                if (AIndex < FCount) then HeapCopy(@FList^[AIndex], @FList^[AIndex + 1], (FCount - AIndex));

                // the list is reallocated by fdelta.
                if (FCount mod FDelta = 0) then SetCapacity(FCount);
        end
        else begin
                {$IFDEF LIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Delete index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TAList.Delete', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAList.Find(AItem: Pointer): Integer;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= 0;
        while (Result < FCount)and(FList^[Result] <> AItem) do
              Inc(Result);

        if (Result = FCount) then Result:= -1;

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.Clear;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        // even if the list is cleared keep an advance
        // in memory allocation to make it faster.

        FCount:= 0;
        FDelta:= 4;

        SetCapacity(FDelta);

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAList.Remove(AItem: Pointer): Integer;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= Find(AItem);
        if (Result > -1) then Delete(Result);

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.Insert(AIndex: Integer; AItem: Pointer);
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then
        begin
                if (FCount = FCapacity) then Grow;

                HeapCopyRev(@FList^[AIndex + 1], @FList^[AIndex], FCount - AIndex);
                FList^[AIndex]:= AItem;
                Inc(FCount);
        end
        else Add(AItem);

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.Exchange(AIndex1, AIndex2: Integer);
var Item: Pointer;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex1 > -1)and(AIndex2 > -1)and(AIndex1 < FCount)and(AIndex2 < FCount) then
        begin
                Item:= FList^[AIndex1];
                FList^[AIndex1]:= FList^[AIndex2];
                FList^[AIndex2]:= Item;
        end
        else begin
                {$IFDEF LIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Exchange index out of bounds. Requested ' + IntToStr(AIndex1) + ' with ' + IntToStr(AIndex2) + ', count ' + IntToStr(FCount) + '.', 'TAList.Exchange', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAList.Move(AIndex, ANewIndex: Integer);
var Item: Pointer;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex <> ANewIndex) then
        begin
                if (AIndex > -1)and(ANewIndex > -1)and
                   (AIndex < FCount)and(ANewIndex < FCount) then
                begin
                        Item:= GetItem(AIndex);

                        Delete(AIndex);
                        Insert(ANewIndex, Item);
                end
                else begin
                        {$IFDEF LIST_THREADSAFE}
                         if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                        {$ENDIF}

                        (*++++++++++++++*)
                        (**) RaiseException('Move index out of bounds. Requested ' + IntToStr(AIndex) + ' to ' + IntToStr(ANewIndex) + ', count ' + IntToStr(FCount) + '.', 'TAList.Move', Self);
                        (*++++++++++++++*)
                end;
        end;

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAList.InBounds(AIndex: Integer): Boolean;
begin
      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= (AIndex > -1)and(AIndex < FCount);

      {$IFDEF LIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;




//##############################################################################
//##                                 STRINGLIST                               ##
//##############################################################################


{ TAStringList }




constructor TAStringList.Create(AAllowLock: Boolean = False);
begin
{$IFDEF STRINGLIST_THREADSAFE}
 if (FAllowLock) then FLock:= TALock.Create('StringList lock');
{$ENDIF}

FAllowLock:= AAllowLock;



        FCount:= 0;
        FCapacity:= 0;
        FDelta:= 4;



        SetCapacity(FDelta);
end;



destructor TAStringList.Destroy;
begin
        SetCapacity(0);



{$IFDEF STRINGLIST_THREADSAFE}
 if (FAllowLock) then FreeNil(FLock);
{$ENDIF}

inherited;
end;


function TAStringList.GetItem(AIndex: Integer): String;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then Result:= FList^[AIndex]
        else begin
                {$IFDEF STRINGLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Getitem index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TAStringList.GetItem', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.SetItem(AIndex: Integer; const AItem: String);
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then FList^[AIndex]:= AItem
        else begin
                {$IFDEF STRINGLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Setitem index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TAStringList.SetItem', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.SetCapacity(ACapacity: Integer);
begin
        if (ACapacity <> FCapacity) then
        begin
                ReallocMem(FList, ACapacity * SizeOf(Pointer));
                FCapacity:= ACapacity;
        end;
end;


procedure TAStringList.Grow;
begin
        // set the list gradularity according
        // to how many items are sored in.

        if (FCapacity > 64) then FDelta:= 32
        else if (FCapacity > 32) then FDelta:= 16
        else FDelta:= 4;

        SetCapacity(FCapacity + FDelta);
end;


procedure TAStringList.Save(const AFileName: String);
var AFileStream: TAFileStream;
    i: Integer;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        AFileStream:= TAFileStream.Create(AFileName, FM_READWRITE, FM_SHARE_NONE, FM_ACCESS_SEQUENTIAL);

        for i:= 0 to FCount - 2 do
        begin
                AFileStream.WriteBuffer(Pointer(FList^[i])^, Length(FList^[i]));
                AFileStream.WriteBuffer(EOL, 2);
        end;
        AFileStream.WriteBuffer(Pointer(FList^[FCount - 1])^, Length(FList^[FCount - 1]));

        FreeNil(AFileStream);

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAStringList.Load(const AFileName: String): Boolean;
var AFileStream: TAFileStream;
    S, Sub: String;
    P, Start: PChar;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if FileExists(AFileName) then
        begin
                Result:= True;
                AFileStream:= TAFileStream.Create(AFileName, FM_READWRITE, FM_SHARE_NONE, FM_ACCESS_SEQUENTIAL);

                SetString(S, nil, AFileStream.Size);
                AFileStream.ReadBuffer(Pointer(S)^, AFileStream.Size);

                FreeNil(AFileStream);


                P:= Pointer(S);

                if (P <> nil) then
                while (P^ <> #0) do
                begin
                        Start:= P;

                        while not (P^ in [#0, #10, #13]) do Inc(P);

                        SetString(Sub, Start, P - Start);
                        Add(Sub);

                        while (P^ in [#10, #13]) do Inc(P);
                end;
        end
        else Result:= False;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAStringList.Add(const AItem: String): Integer;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= FCount;

        // if the list is full then reallocate memory for it.
        if (Result = FCapacity) then Grow;

        FList^[Result]:= AItem;
        Inc(FCount);

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.Delete(AIndex: Integer);
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then
        begin
                // free the memory for the long string which
                // is in fact a dynamic array.
                Finalize(FList^[AIndex]);
                Dec(FCount);

                // if the index is between other items then move the items next to it.
                if (AIndex < FCount) then HeapCopy(@FList^[AIndex], @FList^[AIndex + 1], (FCount - AIndex));

                // the list is reallocated by fdelta.
                if (FCount mod FDelta = 0) then SetCapacity(FCount);
        end
        else begin
                {$IFDEF STRINGLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Delete index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TAStringList.Delete', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAStringList.Find(const AItem: String; ACaseSensitive: Boolean = True): Integer;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= 0;
        while (Result < FCount)and(not StringEqual(FList^[Result], AItem, ACaseSensitive)) do
              Inc(Result);

        if (Result = FCount) then Result:= -1;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.Sort(AStart: Integer = 0; AEnd: Integer = -1);
var i, j, k: Integer;
begin
        if (FCount > 0) then
        begin
                if (AEnd = -1) then AEnd:= FCount - 1;

                repeat
                        i:= AStart;
                        j:= AEnd;
                        k:= (i + j) shr 1;

                        repeat
                                while StringCompare(FList^[i], FList^[k]) < 0 do Inc(i);
                                while StringCompare(FList^[j], FList^[k]) > 0 do Dec(j);

                                if (i <= j) then
                                begin
                                        Exchange(i, j);

                                        if (k = i) then k:= j
                                        else if (k = j) then k:= i;

                                        Inc(i);
                                        Dec(j);
                                end;
                        until (i > j);

                        if (AStart < j) then Sort(AStart, j);
                        AStart:= i;
                until (i >= AEnd);
        end;
end;


procedure TAStringList.Clear;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        // free all the strings in the list which are in fact dynamic arrays.
        Finalize(FList^[0], FCount);

        // even if the list is cleared keep an advance
        // in memory allocation to make it faster.

        FCount:= 0;
        FDelta:= 4;

        SetCapacity(FDelta);

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAStringList.Remove(const AItem: String; ACaseSensitive: Boolean = True): Integer;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= Find(AItem, ACaseSensitive);
        if (Result > -1) then Delete(Result);

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.Insert(AIndex: Integer; const AItem: String);
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then
        begin
                if (FCount = FCapacity) then Grow;

                HeapCopyRev(@FList^[AIndex + 1], @FList^[AIndex], FCount - AIndex);
                FList^[AIndex]:= AItem;
                Inc(FCount);
        end
        else Add(AItem);

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.Exchange(AIndex1, AIndex2: Integer);
var Item: String;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex1 > -1)and(AIndex2 > -1)and(AIndex1 < FCount)and(AIndex2 < FCount) then
        begin
                Item:= FList^[AIndex1];
                FList^[AIndex1]:= FList^[AIndex2];
                FList^[AIndex2]:= Item;
        end
        else begin
                {$IFDEF STRINGLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Exchange index out of bounds. Requested ' + IntToStr(AIndex1) + ' with ' + IntToStr(AIndex2) + ', count ' + IntToStr(FCount) + '.', 'TAStringList.Exchange', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TAStringList.Move(AIndex, ANewIndex: Integer);
var Item: String;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex <> ANewIndex) then
        begin
                if (AIndex > -1)and(ANewIndex > -1)and
                   (AIndex < FCount)and(ANewIndex < FCount) then
                begin
                        Item:= FList^[AIndex];

                        Delete(AIndex);
                        Insert(ANewIndex, Item);
                end
                else begin
                        {$IFDEF STRINGLIST_THREADSAFE}
                         if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                        {$ENDIF}

                        (*++++++++++++++*)
                        (**) RaiseException('Move index out of bounds. Requested ' + IntToStr(AIndex) + ' to ' + IntToStr(ANewIndex) + ', count ' + IntToStr(FCount) + '.', 'TAStringList.Move', Self);
                        (*++++++++++++++*)
                end;
        end;

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TAStringList.InBounds(AIndex: Integer): Boolean;
begin
      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        Result:= (AIndex > -1)and(AIndex < FCount);

      {$IFDEF STRINGLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;




//##############################################################################
//##                                  RECTLIST                                ##
//##############################################################################


{ TARectList }




constructor TARectList.Create(AAllowLock: Boolean = False);
begin
{$IFDEF RECTLIST_THREADSAFE}
 if (FAllowLock) then FLock:= TALock.Create('RectList lock');
{$ENDIF}

FAllowLock:= AAllowLock;



        FCount:= 0;
        FCapacity:= 0;
        FDelta:= 4;



        SetCapacity(FDelta);
end;


destructor TARectList.Destroy;
begin
        SetCapacity(0);



{$IFDEF RECTLIST_THREADSAFE}
 if (FAllowLock) then FreeNil(FLock);
{$ENDIF}

inherited;
end;


function TARectList.GetItem(AIndex: Integer): TRect;
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then Result:= PRect(FList^[AIndex])^
        else begin
                {$IFDEF RECTLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Getitem index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TARectList.GetItem', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TARectList.SetItem(AIndex: Integer; AItem: TRect);
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then PRect(FList^[AIndex])^:= AItem
        else begin
                {$IFDEF RECTLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Setitem index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TARectList.SetItem', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TARectList.SetCapacity(ACapacity: Integer);
begin
        if (ACapacity <> FCapacity) then
        begin
                ReallocMem(FList, ACapacity * SizeOf(Pointer));
                FCapacity:= ACapacity;
        end;
end;


procedure TARectList.Grow;
begin
        // set the list gradularity according
        // to how many items are sored in.

        if (FCapacity > 64) then FDelta:= 32
        else if (FCapacity > 32) then FDelta:= 16
        else FDelta:= 4;

        SetCapacity(FCapacity + FDelta);
end;


procedure TARectList.Add(AItem: TRect);
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        // if the list is full then reallocate memory for it.
        if (FCount = FCapacity) then Grow;


        GetMem(FList^[FCount], SizeOf(TRect));
        PRect(FList^[FCount])^:= AItem;
        Inc(FCount);

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TARectList.Add(AItem: TRect; AMode: Integer);
var i, OldCount: Integer;
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        OldCount:= FCount;

        case AMode of
          RL_MODE_NONE: Add(AItem);
          //
          RL_MODE_AND:
          if (FCount > 0) then
          begin
                  i:= 0;
                  while (i < FCount) do
                  begin
                          if not RestrictIn(PRect(FList^[i])^, AItem) then Delete(i)
                          else Inc(i);
                  end;
          end;
          //
          RL_MODE_OR:
          if (FCount > 0) then
          begin
                  i:= 0;
                  while (i < OldCount)and(i < FCount) do
                  begin
                          if not RestrictOut(PRect(FList^[i])^, AItem) then
                          begin
                                  Dec(OldCount);
                                  Delete(i);
                          end
                          else Inc(i);
                  end;
                  Add(AItem);
          end
          else Add(AItem);
          //
          RL_MODE_DIFF:
          if (FCount > 0) then
          begin
                  i:= 0;
                  while (i < OldCount)and(i < FCount) do
                  begin
                          if not RestrictOut(PRect(FList^[i])^, AItem) then
                          begin
                                  Dec(OldCount);
                                  Delete(i);
                          end
                          else Inc(i);
                  end;
          end
          else Add(AItem);
          //
        end;

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TARectList.Delete(AIndex: Integer);
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        if (AIndex > -1)and(AIndex < FCount) then
        begin
                FreeMem(FList^[AIndex]);
                Dec(FCount);

                // if the index is between other items then move the items next to it.
                if (AIndex < FCount) then HeapCopy(@FList^[AIndex], @FList^[AIndex + 1], (FCount - AIndex));

                // the list is reallocated by fdelta.
                if (FCount mod FDelta = 0) then SetCapacity(FCount);
        end
        else begin
                {$IFDEF RECTLIST_THREADSAFE}
                 if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
                {$ENDIF}

                (*++++++++++++++*)
                (**) RaiseException('Delete index out of bounds. Requested ' + IntToStr(AIndex) + ', count ' + IntToStr(FCount) + '.', 'TARectList.Delete', Self);
                (*++++++++++++++*)
        end;

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TARectList.Clear;
var i: Integer;
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        // free the elements in the list
        for i:= 0 to FCount - 1 do FreeMem(FList^[i]);

        // even if the list is cleared keep an advance
        // in memory allocation to make it faster.

        FCount:= 0;
        FDelta:= 4;

        SetCapacity(FDelta);

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


procedure TARectList.RemoveEmpty;
var i: Integer;
begin
      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.SetExclusive(FLock);
      {$ENDIF}

        i:= 0;
        while (i < FCount) do
        begin
                if RectEmpty(PRect(FList^[i])^) then Delete(i)
                else Inc(i);
        end;

      {$IFDEF RECTLIST_THREADSAFE}
       if (FAllowLock) then ThreadManager.ClearExclusive(FLock);
      {$ENDIF}
end;


function TARectList.RestrictIn(var ARect1: TRect; ARect2: TRect): Boolean;
begin
        Restrict(ARect1, ARect2);
        Result:= not RectEmpty(ARect1);
end;


function TARectList.RestrictOut(var ARect1: TRect; ARect2: TRect): Boolean;
var BTop, BBottom: Integer;
    RetRect: TRect;
begin
        Result:= False;

        if (ARect1.Bottom <= ARect2.Top)or
           (ARect1.Top >= ARect2.Bottom)or
           (ARect1.Right <= ARect2.Left)or
           (ARect1.Left >= ARect2.Right) then
        begin
                Result:= True;
                Exit;
        end;

        if (ARect1.Top >= ARect2.Top) then BTop:= ARect1.Top
        else begin
                BTop:= ARect2.Top;
                if Result then Add(NewRect(ARect1.Left, ARect1.Top, ARect1.Right, BTop))
                else begin
                        RetRect:= NewRect(ARect1.Left, ARect1.Top, ARect1.Right, BTop);
                        Result:= True;
                end;
        end;

        if (ARect1.Bottom <= ARect2.Bottom) then BBottom:= ARect1.Bottom
        else begin
                BBottom:= ARect2.Bottom;
                if Result then Add(NewRect(ARect1.Left, BBottom, ARect1.Right, ARect1.Bottom))
                else begin
                        RetRect:= NewRect(ARect1.Left, BBottom, ARect1.Right, ARect1.Bottom);
                        Result:= True;
                end;
        end;

        if (ARect1.Left < ARect2.Left) then
                if Result then Add(NewRect(ARect1.Left, BTop, ARect2.Left, BBottom))
                else begin
                        RetRect:= NewRect(ARect1.Left, BTop, ARect2.Left, BBottom);
                        Result:=True;
                end;

        if (ARect1.Right > ARect2.Right) then
                if Result then Add(NewRect(ARect2.Right, BTop, ARect1.Right, BBottom))
                else begin
                        RetRect:= NewRect(ARect2.Right, BTop, ARect1.Right, BBottom);
                        Result:= True;
                end;

        if Result then ARect1:= RetRect
        else ARect1:= EMPTY_RECT;
end;




//##############################################################################
//##                                  LOCK                                    ##
//##############################################################################


{ TALock }




constructor TALock.Create(const AName: String);
begin
{$IFDEF EXCEPTION_FROZENLOCK}
 FLastOwnerId:= 0;
 FCounter:= 0;
{$ENDIF}

FName:= AName;



        FLockCount:= 0;
        FOwnerId:= 0;



        ThreadManager.AddLock(Self);

        HWindows.InitializeCriticalSection(FLock);
end;


destructor TALock.Destroy;
begin
        HWindows.DeleteCriticalSection(FLock);

        ThreadManager.RemoveLock(Self);



inherited;
end;




//##############################################################################
//##                               THREAD                                     ##
//##############################################################################


{ TAThread }




constructor TAThread.Create(const AName: String; ACreateSuspended: Boolean; AFreeOnTerminate: Boolean = False; APriority: Integer = TP_NORMAL);
begin
FName:= AName;
FFreeOnTerminate:= AFreeOnTerminate;



        FHandle:= 0;
        FThreadId:= 0;
        FTerminated:= False;
        FFinished:= False;



        ThreadManager.AddThread(Self);

        case ACreateSuspended of
          True:
          begin
                FHandle:= BeginThread(nil, 0, @ThreadProc, Pointer(Self), CREATE_SUSPENDED, FThreadID);
                FSuspendCount:= 1;
          end;
          //
          else begin
                FHandle:= BeginThread(nil, 0, @ThreadProc, Pointer(Self), 0, FThreadID);
                FSuspendCount:= 0;
          end;
          //
        end;

        (*++++++++++++++*)
        (**) if (FHandle = 0) then RaiseException('Cannot create thread "'+ FName + '".', 'TAThread.Create', Self, True);
        (*++++++++++++++*)

        SetPriority(APriority);
end;


destructor TAThread.Destroy;
begin
        if (FHandle <> 0) then
        begin
                if (not FFinished) then
                begin
                        Terminate;

                        WaitFor;
                end;

                HWindows.CloseHandle(FHandle);

                ThreadManager.RemoveThread(Self);
        end;



(*==============*)
(**) if Assigned(FOnDestroy) then FOnDestroy(Self);
(*==============*)

inherited;
end;


function TAThread.GetPriority: Integer;
begin
        Result:= HWindows.GetThreadPriority(FHandle);
end;


procedure TAThread.SetPriority(APriority: Integer);
begin
        HWindows.SetThreadPriority(FHandle, APriority);
end;


function TAThread.GetSuspended: Boolean;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        Result:= (FSuspendCount > 0);

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThread.GetExitCode: Cardinal;
begin
        HWindows.GetExitCodeThread(FHandle, Result);
end;


function TAThread.Suspend: Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        FSuspendCount:= HWindows.SuspendThread(FHandle) + 1;

        Result:= FSuspendCount;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThread.Resume: Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        if (FSuspendCount > 0) then FSuspendCount:= HWindows.ResumeThread(FHandle) - 1;

        Result:= FSuspendCount;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


procedure TAThread.Terminate(AResumeAll: Boolean = True);
begin
        FTerminated:= True;

        // try to resume the thread from the suspended state.
        // check class body to see a warning.
        if (AResumeAll) then while (Resume <> 0) do;
end;


procedure TAThread.WaitFor(AHasMessageQueue: Boolean = False);
var WaitResult: Cardinal;
begin
        if (not FFinished) then
        begin
                (*--------------*)
                (**) if (AHasMessageQueue) then HWindows.PostThreadMessage(FThreadId, WM_QUIT, 0, 0);
                (*--------------*)

                WaitResult:= WaitForSingleObject(FHandle, INFINITE);

                (*++++++++++++++*)
                (**) if (WaitResult = WAIT_FAILED) then RaiseException('Waitfor thread "' + FName + '" failed.', 'TAThread.WaitFor', Self);
                (*++++++++++++++*)
        end;
end;


procedure TAThread.Yield(ATime: Integer = 0);
begin
        HWindows.Sleep(ATime);
end;




//##############################################################################
//##                                THREADMANAGER                             ##
//##############################################################################


{ TAThreadManager }




constructor TAThreadManager.Create;
begin



        FLockList:= TAList.Create;
        FLockCount:= 0;

        FThreadList:= TAList.Create;
        FThreadCount:= 0;
end;


destructor TAThreadManager.Destroy;
begin
        (*++++++++++++++*)
        (**) if (FThreadCount > 0) then RaiseException('All threads should be freed by this time. Count ' + IntToStr(FThreadCount) + '.', 'TAThreadManager.Destroy', Self);
        (*++++++++++++++*)

        FreeNil(FThreadList);

        (*++++++++++++++*)
        (**) if (FLockCount > 0) then RaiseException('All locks should be freed by this time. Count ' + IntToStr(FLockCount) + '.', 'TAThreadManager.Destroy', Self);
        (*++++++++++++++*)

        FreeNil(FLockList);



inherited;
end;


function TAThreadManager.GetLock(AIndex: Integer): TALock;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(LockListLock);
      (*><><><><><><><*)

        if (AIndex > -1)and(AIndex < FLockCount) then Result:= TALock(FLockList.Items[AIndex])
        else begin
                Result:= nil;

                (*><><><><><><><*)
                (**) HWindows.LeaveCriticalSection(LockListLock);
                (*><><><><><><><*)

                (*++++++++++++++*)
                (**) RaiseException('Getlock index out of bounds. Count ' + IntToStr(FLockCount) + '.', 'TAThreadManager.GetLock', Self);
                (*++++++++++++++*)
        end;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(LockListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.GetThread(AIndex: Integer): TAThread;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        if (AIndex > -1)and(AIndex < FThreadCount) then Result:= TAThread(FThreadList.Items[AIndex])
        else begin
                Result:= nil;

                (*><><><><><><><*)
                (**) HWindows.LeaveCriticalSection(ThreadListLock);
                (*><><><><><><><*)

                (*++++++++++++++*)
                (**) RaiseException('Getthread index out of bounds. Count ' + IntToStr(FThreadCount) + '.', 'TAThreadManager.GetThread', Self);
                (*++++++++++++++*)
        end;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.GetRunningThreadId: Cardinal;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        Result:= HWindows.GetCurrentThreadId;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.GetRunningThread: TAThread;
var AIndex: Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        AIndex:= FindThread(HWindows.GetCurrentThreadId);
        if (AIndex > -1) then Result:= FThreadList.Items[AIndex]
        else Result:= nil;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.SetExclusive(ATimes: Integer = 1): Integer;
begin
        Result:= SetExclusive(GlobalLock, ATimes);
end;


function TAThreadManager.SetExclusive(ALock: TALock; ATimes: Integer = 1): Integer;
{$IFDEF EXCEPTION_DEBUG}

var AId: Cardinal;
    AIndex: Integer;
    AThread: TAThread;

{$ENDIF}
begin
        if (ALock <> nil) then
        begin
                if (ATimes > 0) then
                begin
                        {$IFDEF EXCEPTION_DEBUG}

                          (*><><><><><><><*)
                          (**) HWindows.EnterCriticalSection(ThreadListLock);
                          (*><><><><><><><*)

                            AId:= HWindows.GetCurrentThreadId;
                            AIndex:= FindThread(AId);

                            if (AIndex > -1) then
                            begin
                                    AThread:= TAThread(FThreadList.Items[AIndex]);
                                    AThread.FLastLockTry:= ALock.FName;
                                    AThread.FLastLockAction:= L_TRY;
                            end
                            else begin
                                    FLastLockTry:= ALock.FName;
                                    FLastLockAction:= L_TRY;
                            end;

                          (*><><><><><><><*)
                          (**) HWindows.LeaveCriticalSection(ThreadListLock);
                          (*><><><><><><><*)

                        {$ENDIF}



                        HWindows.EnterCriticalSection(ALock.FLock);

                        if (ALock.FLockCount = 0) then ALock.FOwnerId:= HWindows.GetCurrentThreadId;
                        Inc(ALock.FLockCount);
                        Result:= ALock.FLockCount;



                        {$IFDEF EXCEPTION_DEBUG}

                          (*><><><><><><><*)
                          (**) HWindows.EnterCriticalSection(ThreadListLock);
                          (*><><><><><><><*)

                            AId:= HWindows.GetCurrentThreadId;
                            AIndex:= FindThread(AId);

                            if (AIndex > -1) then
                            begin
                                    AThread:= TAThread(FThreadList.Items[AIndex]);
                                    AThread.FLastLockSet:= ALock.FName + ' [' + IntToStr(ALock.FLockCount) + ']';
                                    AThread.FLastLockAction:= L_SET;
                            end
                            else begin
                                    FLastLockSet:= ALock.FName + ' [' + IntToStr(ALock.FLockCount) + ']';
                                    FLastLockAction:= L_SET;
                            end;

                          (*><><><><><><><*)
                          (**) HWindows.LeaveCriticalSection(ThreadListLock);
                          (*><><><><><><><*)

                        {$ENDIF}
                end
                else Result:= 0;

                if (ATimes > 1) then SetExclusive(ALock, ATimes - 1);
        end
        else begin
                Result:= 0;

                (*++++++++++++++*)
                (**) RaiseException('Lock is nil.', 'TAThreadManager.SetExclusive', Self, True);
                (*++++++++++++++*)
        end;
end;


function TAThreadManager.ClearExclusive(ATimes: Integer = 1): Integer;
begin
        Result:= ClearExclusive(GlobalLock, ATimes);
end;


function TAThreadManager.ClearExclusive(ALock: TALock; ATimes: Integer = 1): Integer;
{$IFDEF EXCEPTION_DEBUG}

var AId: Cardinal;
    AIndex: Integer;
    AThread: TAThread;

{$ENDIF}
begin
        if (ALock <> nil) then
        begin
                (*++++++++++++++*)
                (**) if (ALock.FLockCount = 0) then RaiseException('Lock "' + ALock.FName + '" was not set.', 'TAThreadManager.ClearExclusive', Self, True);
                (*++++++++++++++*)

                if (ATimes > 0) then
                begin
                        Result:= ALock.FLockCount;
                        Dec(ALock.FLockCount);
                        if (ALock.FLockCount = 0) then ALock.FOwnerId:= 0;

                        HWindows.LeaveCriticalSection(ALock.FLock);



                        {$IFDEF EXCEPTION_DEBUG}

                          (*><><><><><><><*)
                          (**) HWindows.EnterCriticalSection(ThreadListLock);
                          (*><><><><><><><*)

                            AId:= HWindows.GetCurrentThreadId;
                            AIndex:= FindThread(AId);

                            if (AIndex > -1) then
                            begin
                                    AThread:= TAThread(FThreadList.Items[AIndex]);
                                    AThread.FLastLockClear:= ALock.FName + ' [' + IntToStr(ALock.FLockCount) + ']';
                                    AThread.FLastLockAction:= L_CLEAR;
                            end
                            else begin
                                    FLastLockClear:= ALock.FName + ' [' + IntToStr(ALock.FLockCount) + ']';
                                    FLastLockAction:= L_CLEAR;
                            end;

                          (*><><><><><><><*)
                          (**) HWindows.LeaveCriticalSection(ThreadListLock);
                          (*><><><><><><><*)

                        {$ENDIF}
                end
                else Result:= 0;

                if (ATimes > 1) then ClearExclusive(ALock, ATimes - 1);
        end
        else begin
                Result:= 0;

                (*++++++++++++++*)
                (**) RaiseException('Lock is nil.', 'TAThreadManager.ClearExclusive', Self, True)
                (*++++++++++++++*)
        end;
end;


function TAThreadManager.AddThread(AThread: TAThread): Boolean;
var AIndex: Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        if (AThread <> nil) then
        begin
                AIndex:= FindThread(AThread.FName);

                if (AIndex = -1) then
                begin
                        FThreadList.Add(AThread);
                        Inc(FThreadCount);
                        Result:= True;
                end
                else begin
                        Result:= False;

                        (*><><><><><><><*)
                        (**) HWindows.LeaveCriticalSection(ThreadListLock);
                        (*><><><><><><><*)

                        (*++++++++++++++*)
                        (**) RaiseException('Thread name "' + AThread.Name + '" not unique.', 'TAThreadManager.AddThread', Self, True)
                        (*++++++++++++++*)
                end;
        end
        else Result:= False;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.RemoveThread(AThread: TAThread): Boolean;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        if (AThread <> nil) then
        begin
                FThreadList.Remove(AThread);
                Dec(FThreadCount);
                Result:= True;
        end
        else Result:= False;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.FindThread(AThread: TAThread): Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        Result:= 0;
        while (Result < FThreadCount)and(TAThread(FThreadList.Items[Result]) <> AThread) do
              Inc(Result);

        if (Result = FThreadCount) then Result:= -1;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.FindThread(AThreadId: Cardinal): Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        Result:= 0;
        while (Result < FThreadCount)and(TAThread(FThreadList.Items[Result]).FThreadId <> AThreadId) do
              Inc(Result);

        if (Result = FThreadCount) then Result:= -1;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.FindThread(const AName: String): Integer;
begin
      (*><><><><><><><*)
      (**) HWindows.EnterCriticalSection(ThreadListLock);
      (*><><><><><><><*)

        Result:= 0;
        while (Result < FThreadCount)and(not StringEqual(TAThread(FThreadList.Items[Result]).FName, AName, False)) do
              Inc(Result);

        if (Result = FThreadCount) then Result:= -1;

      (*><><><><><><><*)
      (**) HWindows.LeaveCriticalSection(ThreadListLock);
      (*><><><><><><><*)
end;


function TAThreadManager.AddLock(ALock: TALock): Boolean;
begin
        (*><><><><><><><*)
        (**) HWindows.EnterCriticalSection(LockListLock);
        (*><><><><><><><*)

          FLockList.Add(ALock);
          Inc(FLockCount);
          Result:= True;

        (*><><><><><><><*)
        (**) HWindows.LeaveCriticalSection(LockListLock);
        (*><><><><><><><*)
end;


function TAThreadManager.RemoveLock(ALock: TALock): Boolean;
begin
        (*><><><><><><><*)
        (**) HWindows.EnterCriticalSection(LockListLock);
        (*><><><><><><><*)

          (*++++++++++++++*)
          (**) if (ALock.OwnerId <> 0) then RaiseException('Lock "' + ALock.FName + '" should be cleared before removal.', 'TAThreadManager.RemoveLock', Self);
          (*++++++++++++++*)

          FLockList.Remove(ALock);
          Dec(FLockCount);
          Result:= True;

        (*><><><><><><><*)
        (**) HWindows.LeaveCriticalSection(LockListLock);
        (*><><><><><><><*)
end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//###############################              #################################
//###############################              #################################
//##                               EXCEPTIONS                                 ##
//###############################              #################################
//###############################              #################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                                 EXCEPTION                                ##
//##############################################################################


{ Exception }




constructor Exception.Create(const AMessage, ALocation: String; ARaiser: TObject; AFatal: Boolean = False);
begin
FMessage:= AMessage;
FLocation:= ALocation;
FRaiser:= ARaiser;
FFatal:= AFatal;



end;




//##############################################################################
//##                                HEAPEXCEPTION                             ##
//##############################################################################


{ HeapException }




procedure EHeapException.FreeInstance;
begin



if FAllowFree then inherited FreeInstance;
end;



//##############################################################################
//##                          exception handling                              ##
//##############################################################################

procedure RaiseException(const AMessage, ALocation: String; ARaiser: TObject; AFatal: Boolean = False);
var AException: Exception;
begin
          AException:= Exception.Create(AMessage, ALocation, ARaiser, AFatal);

          raise AException;
end;


procedure RaiseLastError(const ALocation: String; ARaiser: TObject; AFatal: Boolean = False);
var ALastError: Integer;
begin
        ALastError:= HWindows.GetLastError;

        if (ALastError <> 0) then
                RaiseException(ErrorMessage(ALastError), ALocation, ARaiser, AFatal);
end;


procedure HandleException(const ALocation: String);
var AException: Exception;
    Msg, Debug: String;
begin
        (*><><><><><><><*)
        (**) HWindows.EnterCriticalSection(ExceptionLock);
        (*><><><><><><><*)

          AException:= Exception(ExceptObject);
          Msg:= ExceptionMessage(AException, ALocation);
          Debug:= DebugMessage(AException, ALocation);

          {$IFDEF EXCEPTION_LOG}
            ExceptionLog(Debug);
          {$ENDIF}

          {$IFDEF EXCEPTION_SHOWWINDOW}
            if (AException <> nil) then ExceptionFatal:= AException.Fatal
            else ExceptionFatal:= True;


            // show window and wait for it to close. hang in a lock so other threads
            // will have to wait in line for an exception raise.
            SetWindowText(ExceptionMsgHandle, PChar(Msg));
            SetWindowText(ExceptionDebugHandle, PChar(Debug));
            SetWindowPos(ExceptionWindowHandle, 0, 0, 0, 0, 0, SWP_SHOWWINDOW or SWP_NOMOVE or SWP_NOSIZE or SWP_ASYNCWINDOWPOS);
            ExceptionVisible:= True;


            while ExceptionVisible do Sleep(10);
          {$ELSE}
            if AException.Fatal then Halt(0);
          {$ENDIF}

        (*><><><><><><><*)
        (**) HWindows.LeaveCriticalSection(ExceptionLock);
        (*><><><><><><><*)
end;


function ExceptionMessage(AException: Exception; const AHandleLocation: String): String;
begin
        if (AException <> nil) then
        begin
                if (AException.Raiser <> nil) then Result:= 'Exception with message "' + AException.Message + '" raised by raiser ' + AException.Raiser.ClassName + ' at location "' + AException.Location + '". Handled at location "' + AHandleLocation +'".' + EOL + EOL
                else Result:= 'Exception with message "' + AException.Message + '" raised by N/A at location "' + AException.Location + '". Handled at location "' + AHandleLocation +'".' + EOL + EOL
        end
        else Result:= 'Exception raised but exception object not avaible.' + EOL + EOL;
end;


function DebugMessage(AException: Exception; const AHandleLocation: String): String;
{$IFDEF EXCEPTION_DEBUG}

var ALock: TALock;
    AThread: TAThread;
    AId: Cardinal;
    AIndex, i, j: Integer;

{$ENDIF}
begin
        {$IFDEF EXCEPTION_DEBUG}

          (*><><><><><><><*)
          (**) HWindows.EnterCriticalSection(LockListLock);
          (**) HWindows.EnterCriticalSection(ThreadListLock);
          (*><><><><><><><*)

            AId:= HWindows.GetCurrentThreadId;

            Result:= 'FileName: "'+ ParamStr(0) + '"' + EOL;
            Result:= Result + 'Memory: '+ IntToStr(AHeap.HeapSize) + EOL + EOL;

            if (AException.Raiser <> nil) then Result:= Result + 'Exception class: ' + AException.Raiser.ClassName + EOL
            else Result:= Result + 'Exception class: N/A' + EOL;
            Result:= Result + 'Exception adress: $' + IntToHex(Integer(AException.Raiser)) + EOL;
            Result:= Result + 'Exception message: "' + AException.Message + '"' + EOL;
            Result:= Result + 'Exception raise location: "'+ AException.Location + '"' + EOL;
            Result:= Result + 'Exception catch location: "'+ AHandleLocation + '"' + EOL;
            Result:= Result + 'Exception raise threadId: '+ CardToStr(AId) + EOL;

            if (ThreadManager <> nil) then
            begin
                // get raiser thread name.
                Result:= Result + 'Exception raise thread: ';
                if (AId = MainThreadId) then Result:= Result + 'Main thread' + EOL + EOL
                else begin
                        AIndex:= ThreadManager.FindThread(HWindows.GetCurrentThreadId);

                        if (AIndex > -1) then Result:= Result + ThreadManager.Threads[AIndex].Name + EOL + EOL
                        else Result:= Result + 'N/A' + EOL + EOL;
                end;

                // add all lock information.
                Result:= Result + 'Module locks:' + EOL + EOL;
                for i:= 0 to ThreadManager.FLockCount-1 do
                begin
                        ALock:= TALock(ThreadManager.FLockList.Items[i]);

                        Result:= Result + ALock.Name + EOL;
                        Result:= Result + '       | ' + 'Owned by thread: ';
                        if (ALock.OwnerId = MainThreadId) then Result:= Result + 'Main thread' + EOL
                        else begin
                                AIndex:= ThreadManager.FindThread(ALock.OwnerId);

                                if (AIndex > -1) then Result:= Result + ThreadManager.Threads[AIndex].Name + EOL
                                else Result:= Result + 'N/A' + EOL;
                        end;
                        Result:= Result + '       | ' + 'Lock count: ' + IntToStr(ALock.LockCount) + EOL;

                        // make exception fatal if a lock is entered.
                        if (ThreadManager.FThreadCount > 0)and(ALock.LockCount > 0) then AException.Fatal:= True;
                end;

                // add all thread information.
                Result:= Result + EOL + 'Module threads:' + EOL + EOL;
                // add main thread information.
                Result:= Result + 'Main thread' + EOL;
                Result:= Result + '       | ' + 'Id: ' + CardToStr(MainThreadId) + EOL;
                Result:= Result + '       | ' + 'Owned locks: ' + EOL;
                for j:= 0 to ThreadManager.FLockCount-1 do
                begin
                        ALock:= TALock(ThreadManager.FLockList.Items[j]);

                        if (ALock.OwnerId = MainThreadId) then Result:= Result + '       |        | ' + ALock.Name + ' [ ' + IntToStr(ALock.LockCount) + ' times ]' + EOL;
                end;
                Result:= Result + '       | ' + 'Last lock set: ' + ThreadManager.FLastLockSet + EOL;
                Result:= Result + '       | ' + 'Last lock clear: ' + ThreadManager.FLastLockClear + EOL;
                Result:= Result + '       | ' + 'Last lock try: ' + ThreadManager.FLastLockTry + EOL;
                Result:= Result + '       | ' + 'Last lock action: ' + ThreadManager.FLastLockAction + EOL;
                // add secondary threads information.
                for i:= 0 to ThreadManager.FThreadCount-1 do
                begin
                        AThread:= TAThread(ThreadManager.FThreadList.Items[i]);

                        Result:= Result + AThread.Name + EOL;
                        Result:= Result + '       | ' + 'Id: ' + CardToStr(AThread.ThreadId) + EOL;
                        Result:= Result + '       | ' + 'FreeOnTerminate: ' + BoolToStr(AThread.FreeOnTerminate) + EOL;
                        Result:= Result + '       | ' + 'State: ' ;
                        if AThread.Finished then Result:= Result + 'Finished' + EOL
                        else if AThread.Terminated then Result:= Result + 'Terminated' + EOL
                        else Result:= Result + 'Running' + EOL;
                        Result:= Result + '       | ' + 'Suspend count: ' + IntToStr(AThread.SuspendCount) + EOL;
                        Result:= Result + '       | ' + 'Owned locks: ' + EOL;
                        for j:= 0 to ThreadManager.FLockCount-1 do
                        begin
                                ALock:= TALock(ThreadManager.FLockList.Items[j]);

                                if (ALock.OwnerId = AThread.ThreadId) then Result:= Result + '       |        | ' + ALock.Name + ' [ ' + IntToStr(ALock.LockCount) + ' times ]' + EOL;
                        end;
                        Result:= Result + '       | ' + 'Last lock set: ' + AThread.FLastLockSet + EOL;
                        Result:= Result + '       | ' + 'Last lock clear: ' + AThread.FLastLockClear + EOL;
                        Result:= Result + '       | ' + 'Last lock try: ' + AThread.FLastLockTry + EOL;
                        Result:= Result + '       | ' + 'Last lock action: ' + AThread.FLastLockAction + EOL;
                end;

                Result:= Result + EOL + EOL;
            end
            else Result:= Result + EOL + 'Threadmanager not yet created. Cannot get thread or lock info.' + EOL + EOL;

          (*><><><><><><><*)
          (**) HWindows.LeaveCriticalSection(ThreadListLock);
          (**) HWindows.LeaveCriticalSection(LockListLock);
          (*><><><><><><><*)

        {$ENDIF}
end;


function ErrorMessage(AErrorCode: Integer): String;
var Len: Integer;
    Buffer: array[0..MAX_PATH - 1] of Char;
begin
        Len:= FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ARGUMENT_ARRAY, nil,
                            AErrorCode, 0, Buffer, SizeOf(Buffer), nil);

        SetString(Result, Buffer, Len);
end;


procedure ExceptionHandler(ExceptObject: TObject; ExceptAddr: Pointer);
begin
        HandleException('Default exception handler');
end;


procedure ErrorHandler(ErrorCode: Byte; ErrorAddr: Pointer);
var AException: Exception;
begin
        case ErrorCode of
          FE_OUTOFMEMORY:               AException:= OutOfMemory;
          FE_INVALIDPOINTER:            AException:= InvalidPointer;
          FE_DIVBYZERO..FE_SAFECALL:    AException:= Exception.Create(EMessages[ErrorCode], 'Runtime error handler', nil, True);
          else                          AException:= Exception.Create('I/O exception.', 'Runtime error handler', nil, True);
        end;


        raise AException at ErrorAddr;
end;


procedure AbstractErrorHandler;
begin
        RaiseException('Abstract error.', 'Abstract error handler', nil, True);
end;

//##############################################################################
//##                      hidden exception functions                          ##
//##############################################################################

function GetExceptionClass(P: PExceptionRecord): ExceptClass;
begin
        Result:= Exception;
end;


function GetExceptionObject(P: PExceptionRecord): Exception;
var ErrorCode: Integer;

  function CreateAVObject: Exception;
  var AccessOp: String;
      AccessAddress: Pointer;
  begin
        with P^ do
        begin
                if ExceptionInformation[0] = 0 then AccessOp:= 'Read'
                else AccessOp:= 'Write';

                AccessAddress:= Pointer(ExceptionInformation[1]);

                Result:= Exception.Create('Access violation. ' + AccessOp + ' of address $'+
                                           IntToHex(Integer(AccessAddress)) + '.', 'GetExceptionObject', nil, True);
        end;
  end;

begin
        // map an OS exception to a delphi class instance.
        ErrorCode:= MapException(P);

        case ErrorCode of
          FE_OUTOFMEMORY:                       Result:= OutOfMemory;
          FE_INVALIDPOINTER:                    Result:= InvalidPointer;
          FE_DIVBYZERO..FE_INVALIDCAST:         Result:= Exception.Create(EMessages[ErrorCode], 'GetExceptionObject', nil, True);
          FE_ACCESSVIOLATION:                   Result:= CreateAVObject;
          FE_PRIVILEGE..FE_ASSERTION:           Result:= Exception.Create(EMessages[ErrorCode], 'GetExceptionObject', nil);
          // external exceptions.
          else                                  Result:= Exception.Create('External exception with code ' + IntToStr(P.ExceptionCode), 'GetExceptionObject', nil, True);
        end;
end;


function MapException(P: PExceptionRecord): Integer;
begin
        case P.ExceptionCode of
          STATUS_INTEGER_DIVIDE_BY_ZERO:      Result:= 3;
          STATUS_ARRAY_BOUNDS_EXCEEDED:       Result:= 4;
          STATUS_INTEGER_OVERFLOW:            Result:= 5;
          STATUS_FLOAT_INEXACT_RESULT,
          STATUS_FLOAT_INVALID_OPERATION,
          STATUS_FLOAT_STACK_CHECK:           Result:= 6;
          STATUS_FLOAT_DIVIDE_BY_ZERO:        Result:= 7;
          STATUS_FLOAT_OVERFLOW:              Result:= 8;
          STATUS_FLOAT_UNDERFLOW,
          STATUS_FLOAT_DENORMAL_OPERAND:      Result:= 9;
          STATUS_ACCESS_VIOLATION:            Result:= 11;
          STATUS_PRIVILEGED_INSTRUCTION:      Result:= 12;
          STATUS_CONTROL_C_EXIT:              Result:= 13;
          STATUS_STACK_OVERFLOW:              Result:= 14;
          //
          else Result:= 22;
        end;
end;

//##############################################################################
//##                        exception log functions                           ##
//##############################################################################

procedure ExceptionLog(const AMessage: String);
begin
        {$IFDEF EXCEPTION_LOG}
          ExceptionLogFile.Write(Pointer(AMessage)^, Length(AMessage));
        {$ENDIF}
end;

//##############################################################################
//##                      exception window functions                          ##
//##############################################################################

function ExceptionThread(APointer: Pointer): Cardinal;
{$IFDEF EXCEPTION_SHOWWINDOW}

var AMsg: TMsg;
    Font: Cardinal;

{$ENDIF}
begin
      {$IFDEF EXCEPTION_SHOWWINDOW}

        // create the window.
        ExceptionWindowClass.hInstance:= MainInstance;
        HWindows.RegisterClass(ExceptionWindowClass);

        ExceptionWindowHandle:= HWindows.CreateWindowEx(
                          WS_EX_TOPMOST,
                          ExceptionWindowClass.lpszClassName, 'Exception catch',
                          WS_CAPTION or WS_BORDER or WS_CLIPCHILDREN,
                          20, 20, 400, 130,
                          0, 0,
                          HInstance, nil);

        (*++++++++++++++*)
        (**) if (ExceptionWindowHandle = 0) then RaiseException('Cannot create handle for exception window.', 'ExceptionThread', nil, True);
        (*++++++++++++++*)

        // now we set the real WindowProc.
        HWindows.SetWindowLong(ExceptionWindowHandle, GWL_WNDPROC, Integer(@ExceptionWindow));

        // set the default cursor.
        HWindows.SetCursor(LoadCursor(0, IDC_ARROW));

        // get the dc.
        ExceptionWindowDC:= HWindows.GetDC(ExceptionWindowHandle);
        ExceptionCloseRect:= NewRect(280, 10, 387, 32);
        ExceptionDebugRect:= NewRect(280, 35, 387, 57);
        Font:= HWindows.CreateFont(14, 0, 0, 0, 100, 0, 0, 0, DEFAULT_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, FF_DONTCARE, 'Arial');
        HWindows.SelectObject(ExceptionWindowDC, Font);

        // set the window to recive hover and leave notifications.
        ExceptionMouseTrack.cbSize:= SizeOf(TTrackMouseEvent);
        ExceptionMouseTrack.dwFlags:= TME_LEAVE;
        ExceptionMouseTrack.hwndTrack:= ExceptionWindowHandle;
        ExceptionMouseTrack.dwHoverTime:= HOVER_DEFAULT;
        HWindows.TrackMouseEvent(ExceptionMouseTrack);

        ExceptionMsgHandle:= HWindows.CreateWindow('EDIT', '', WS_VISIBLE or WS_CHILD or ES_MULTILINE or ES_READONLY, 10, 10, 260, 90, ExceptionWindowHandle, 0, MainInstance, nil);
        HWindows.SendMessage(ExceptionMsgHandle, WM_SETFONT, Font, Integer(True));
        ExceptionDebugHandle:= HWindows.CreateWindow('EDIT', '', WS_VISIBLE or WS_CHILD or WS_HSCROLL or WS_VSCROLL or ES_MULTILINE or ES_READONLY, 10, 125, 365, 250, ExceptionWindowHandle, 0, MainInstance, nil);
        HWindows.SendMessage(ExceptionDebugHandle, WM_SETFONT, Font, Integer(True));



        try
          repeat
               case Integer(HWindows.GetMessage(AMsg, 0, 0, 0)) of
                 //
                 GM_SUCCESS:
                 begin
                        HWindows.TranslateMessage(AMsg);
                        HWindows.DispatchMessage(AMsg);
                 end;
                 //
                 GM_TERMINATE: ExceptionThreadTerminated:= True;
                 //
                 GM_ERROR:
                        (*++++++++++++++*)
                        (**) RaiseException('Cannot getmessage.', 'ExceptionThread', nil);
                        (*++++++++++++++*)
                  //
               end;
          until ExceptionThreadTerminated;
        except
          (*+++++++++*)
          (**) Halt(0);
          (*+++++++++*)
        end;



        // destroy the window.
        HWindows.DestroyWindow(ExceptionMsgHandle);
        HWindows.DestroyWindow(ExceptionDebugHandle);
        HWindows.ReleaseDC(ExceptionWindowHandle, ExceptionWindowDC);
        HWindows.DestroyWindow(ExceptionWindowHandle);
        HWindows.DeleteObject(Font);

        HWindows.GetExitCodeThread(ExceptionThreadId, Result);
        ExceptionThreadFinished:= True;
        HWindows.ExitThread(Result);

      {$ELSE}
        Result:= 0;
      {$ENDIF}
end;


function ExceptionWindow(AHandle: Cardinal; AMsg, WParam, LParam: Integer): Integer; stdcall;
{$IFDEF EXCEPTION_SHOWWINDOW}

var PaintStruct: TPaintStruct;
    Brush: Cardinal;
    Pen: Cardinal;
    State: Integer;
    Text: String;
    Rect: TRect;

{$ENDIF}
begin
      {$IFDEF EXCEPTION_SHOWWINDOW}

        case AMsg of
          WM_PAINT:
          begin
                BeginPaint(AHandle, PaintStruct);

                // pass to validate the area.
                Result:= HWindows.DefWindowProc(AHandle, AMsg, WParam, LParam);


                SetBkMode(ExceptionWindowDC, OPAQUE);

                Brush:= CreateSolidBrush(GetSysColor(COLOR_3DFACE));
                FillRect(ExceptionWindowDC, NewRect(0, 0, 400, 420), Brush);
                DeleteObject(Brush);

                Pen:= CreateSolidBrush(ButtonBorders[ExceptionCloseState]);
                Brush:= CreateSolidBrush(ButtonColors[ExceptionCloseState]);
                FillRect(ExceptionWindowDC, ExceptionCloseRect, Brush);
                FrameRect(ExceptionWindowDC, ExceptionCloseRect, Pen);
                DeleteObject(Brush);
                DeleteObject(Pen);

                {$IFDEF EXCEPTION_DEBUG}
                  Pen:= CreateSolidBrush(ButtonBorders[ExceptionDebugState]);
                  Brush:= CreateSolidBrush(ButtonColors[ExceptionDebugState]);
                  FillRect(ExceptionWindowDC, ExceptionDebugRect, Brush);
                  FrameRect(ExceptionWindowDC, ExceptionDebugRect, Pen);
                  DeleteObject(Brush);
                  DeleteObject(Pen);
                {$ENDIF}

                SetBkMode(ExceptionWindowDC, TRANSPARENT);

                Text:= Texts[Integer(ExceptionFatal)];
                SetTextColor(ExceptionWindowDC, TextColors[ExceptionCloseState]);
                TextOut(ExceptionWindowDC, ExceptionCloseRect.Left + 13, ExceptionCloseRect.Top + 4, PChar(Text), Length(Text));

                {$IFDEF EXCEPTION_DEBUG}
                  SetTextColor(ExceptionWindowDC, TextColors[ExceptionDebugState]);
                  TextOut(ExceptionWindowDC, ExceptionDebugRect.Left + 27, ExceptionDebugRect.Top + 4, 'Debug Log', Length('Debug log'));
                {$ENDIF}

                EndPaint(AHandle, PaintStruct);

                Exit;
          end;
          //
          WM_MOUSEMOVE, WM_LBUTTONDOWN, WM_LBUTTONUP, WM_MOUSELEAVE:
          begin
                if (AMsg = WM_MOUSEMOVE) then
                begin
                        HWindows.SetCursor(LoadCursor(0, IDC_ARROW));
                        ExceptionMouseTrack.cbSize:= SizeOf(TTrackMouseEvent);
                        ExceptionMouseTrack.dwFlags:= TME_LEAVE;
                        ExceptionMouseTrack.hwndTrack:= ExceptionWindowHandle;
                        ExceptionMouseTrack.dwHoverTime:= HOVER_DEFAULT;
                        HWindows.TrackMouseEvent(ExceptionMouseTrack);
                end;

                State:= Integer(PointInRect(LowInt(LParam), HighInt(LParam), ExceptionCloseRect));
                if (State = BS_HOVER)and(GetKeyState(VK_LBUTTON) < 0) then State:= BS_DOWN;

                if (State <> ExceptionCloseState) then
                begin
                        if (State = BS_HOVER)and(ExceptionCloseState = BS_DOWN) then
                        begin
                                if (not ExceptionFatal) then
                                begin
                                        SetWindowPos(ExceptionWindowHandle, 0, 0, 0, ExceptionWindowRect.Right, ExceptionWindowRect.Bottom, SWP_ASYNCWINDOWPOS or SWP_NOZORDER or SWP_NOACTIVATE or SWP_NOMOVE or SWP_HIDEWINDOW);
                                        ExceptionVisible:= False;
                                end
                                else Halt(0);
                        end;

                        ExceptionCloseState:= State;
                        Rect:= NewRect(ExceptionCloseRect.Left -1, ExceptionCloseRect.Top -1, ExceptionCloseRect.Right +1, ExceptionCloseRect.Bottom +1);
                        InvalidateRect(ExceptionWindowHandle, @Rect, False);
                end;

                {$IFDEF EXCEPTION_DEBUG}
                State:= Integer(PointInRect(LowInt(LParam), HighInt(LParam), ExceptionDebugRect));
                if (State = BS_HOVER)and(GetKeyState(VK_LBUTTON) < 0) then State:= BS_DOWN;

                  if (State <> ExceptionDebugState) then
                  begin
                        if (State = BS_HOVER)and(ExceptionDebugState = BS_DOWN) then
                                SetWindowPos(ExceptionWindowHandle, 0, 0, 0, 400, 420, SWP_ASYNCWINDOWPOS or SWP_NOZORDER or SWP_NOACTIVATE or SWP_NOMOVE);

                        ExceptionDebugState:= State;
                        Rect:= NewRect(ExceptionDebugRect.Left -1, ExceptionDebugRect.Top -1, ExceptionDebugRect.Right +1, ExceptionDebugRect.Bottom +1);
                        InvalidateRect(ExceptionWindowHandle, @Rect, False);
                  end;
                {$ENDIF}
          end;
          //
          WM_CLOSE:
          begin
                Result:= 0;
                Exit;
          end;
          //
        end;


        // always pass the message to the default
        // window procedure (system handled).
        Result:= HWindows.DefWindowProc(ExceptionWindowHandle, AMsg, WParam, LParam);

      {$ELSE}
        Result:= 0;
      {$ENDIF}
end;


//##############################################################################
//##                         frozen window functions                          ##
//##############################################################################

function FrozenThread(APointer: Pointer): Cardinal;
{$IFDEF EXCEPTION_FROZENLOCK}

var ALock: TALock;
    i: Integer;

{$ENDIF}
begin
      {$IFDEF EXCEPTION_FROZENLOCK}

        try
          repeat
              (*><><><><><><><*)
              (**) HWindows.EnterCriticalSection(LockListLock);
              (*><><><><><><><*)

                // only run the checker if the exception window is not shown,
                // blocking a lock and if there is more then one thread in the app.
                
                if (ThreadManager.FThreadCount > 0)and(not ExceptionVisible) then
                for i:= 0 to ThreadManager.FLockCount-1 do
                begin
                        ALock:= ThreadManager.FLockList.Items[i];

                        if (ALock.OwnerId <> 0) then
                        begin
                                if (ALock.OwnerId = ALock.FLastOwnerId) then
                                begin
                                        if (ALock.FCounter = 2000) then
                                        begin
                                                try
                                                  (*++++++++++++++*)
                                                  (**) RaiseException('One or more threads are in a deadlock. Lock "' + ALock.FName + '" seems to be one of them.', 'ExceptionFrozen', nil, True);
                                                  (*++++++++++++++*)
                                                except
                                                  (*++++++++++++++*)
                                                  (**) HandleException('ExceptionFrozen');
                                                  (*++++++++++++++*)
                                                end;
                                        end;

                                        Inc(ALock.FCounter);
                                end
                                else ALock.FCounter:= 0;

                                ALock.FLastOwnerId:= ALock.OwnerId;
                        end
                        else begin
                                ALock.FLastOwnerId:= 0;
                                ALock.FCounter:= 0;
                        end;
                end;

              (*><><><><><><><*)
              (**) HWindows.LeaveCriticalSection(LockListLock);
              (*><><><><><><><*)

              HWindows.Sleep(10);

          until FrozenThreadTerminated;
        except
          (*+++++++++*)
          (**) Halt(0);
          (*+++++++++*)
        end;



        HWindows.GetExitCodeThread(FrozenThreadId, Result);
        FrozenThreadFinished:= True;
        HWindows.ExitThread(Result);

      {$ELSE}
        Result:= 0;
      {$ENDIF}
end;

//##############################################################################
//##                        exceptions init and uninit                        ##
//##############################################################################

procedure CreateExceptions;
begin
        ErrorProc:= @ErrorHandler;
        ExceptProc:= @ExceptionHandler;
        AbstractErrorProc:= @AbstractErrorHandler;

        ExceptionClass:= Exception;

        ExceptClsProc:= @GetExceptionClass;
        ExceptObjProc:= @GetExceptionObject;

        OutOfMemory:= EHeapException.Create('Out of memory.', 'N/A', nil, True);
        InvalidPointer:= EHeapException.Create('Invalid pointer operation.', 'N/A', nil, True);
        OutOfMemory.FAllowFree:= False;
        InvalidPointer.FAllowFree:= False;

        HWindows.InitializeCriticalSection(ExceptionLock);
end;


procedure DestroyExceptions;
begin
        HWindows.DeleteCriticalSection(ExceptionLock);

        OutOfMemory.FAllowFree:= True;
        InvalidPointer.FAllowFree:= True;
        FreeNil(OutOfMemory);
        FreeNil(InvalidPointer);
end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#######################                                 ######################
//#######################                                 ######################
//##                              INITIALIZATION                              ##
//#######################                                 ######################
//#######################                                 ######################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
procedure Init;
begin
        CreateExceptions;

        try
          HWindows.InitializeCriticalSection(LockListLock);
          HWindows.InitializeCriticalSection(ThreadListLock);

          ThreadManager:= TAThreadManager.Create;

          GlobalLock:= TALock.Create('Global lock');
          FreeOnTerminateLock:= TALock.Create('FreeOnTerminate lock');



          {$IFDEF EXCEPTION_LOG}
            ExceptionLogFile:= TAFileStream.Create('Exception.Log', FM_WRITE or FM_EMPTY, FM_SHARE_READ, FM_ACCESS_FLUSH);
          {$ENDIF}

          {$IFDEF EXCEPTION_SHOWWINDOW}
            ExceptionThreadHandle:= BeginThread(nil, 0, @ExceptionThread, nil, 0, ExceptionThreadId);
          {$ENDIF}

          {$IFDEF EXCEPTION_FROZENLOCK}
            FrozenThreadHandle:= BeginThread(nil, 0, @FrozenThread, nil, 0, FrozenThreadId);
          {$ENDIF}
        except
          (*++++++++++++++*)
          (**) HandleException('ARoutine intialization');
          (*++++++++++++++*)
        end;
end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#######################                                 ######################
//#######################                                 ######################
//##                                FINALIZATION                              ##
//#######################                                 ######################
//#######################                                 ######################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
procedure Final;
begin
        try
          {$IFDEF EXCEPTION_FROZENLOCK}
            FrozenThreadTerminated:= True;
            while (not FrozenThreadFinished) do Sleep(10);
          {$ENDIF}

          {$IFDEF EXCEPTION_SHOWWINDOW}
            HWindows.PostMessage(ExceptionWindowHandle, WM_QUIT, 0, 0);
            while (not ExceptionThreadFinished) do Sleep(10);
          {$ENDIF}

          {$IFDEF EXCEPTION_LOG}
            FreeNil(ExceptionLogFile);
          {$ENDIF}



          FreeNil(FreeOnTerminateLock);
          FreeNil(GlobalLock);

          FreeNil(ThreadManager);

          HWindows.DeleteCriticalSection(ThreadListLock);
          HWindows.DeleteCriticalSection(LockListLock);
        except
          (*++++++++++++++*)
          (**) HandleException('ARoutine finalization');
          (*++++++++++++++*)
        end;

        DestroyExceptions;
end;

//##############################################################################
//##############################################################################

End.

